import pytest

from diofant import (Derivative, Dummy, Float, I, O, Piecewise, Rational,
                     RisingFactorial, Sum, Tuple, besseli, cos, exp, exp_polar,
                     expand_func, factorial, false, gamma, hyper, limit, log,
                     meijerg, oo, pi, polar_lift, sqrt, symbols)
from diofant.abc import a, b, c, d, k, l, s, x, z
from diofant.functions.special.hyper import (HyperRep, HyperRep_asin1,
                                             HyperRep_asin2, HyperRep_atanh,
                                             HyperRep_cosasin, HyperRep_log1,
                                             HyperRep_log2, HyperRep_power1,
                                             HyperRep_power2, HyperRep_sinasin,
                                             HyperRep_sqrts1, HyperRep_sqrts2)
from diofant.utilities.randtest import random_complex_number as randcplx
from diofant.utilities.randtest import verify_derivative_numerically as td
from diofant.utilities.randtest import verify_numerically as tn


__all__ = ()


def test_TupleParametersBase():
    # test that our implementation of the chain rule works
    p = hyper((), (), z**2)
    assert p.diff(z) == p*2*z


def test_hyper():
    pytest.raises(TypeError, lambda: hyper(1, 2, z))

    assert hyper((1, 2), (1,), z) == hyper(Tuple(1, 2), Tuple(1), z)

    h = hyper((1, 2), (3, 4, 5), z)
    assert h.ap == Tuple(1, 2)
    assert h.bq == Tuple(3, 4, 5)
    assert h.argument == z
    assert h.is_commutative is True

    # just a few checks to make sure that all arguments go where they should
    assert tn(hyper(Tuple(), Tuple(), z), exp(z), z)
    assert tn(z*hyper((1, 1), Tuple(2), -z), log(1 + z), z)

    # differentiation
    h = hyper(
        (randcplx(), randcplx(), randcplx()), (randcplx(), randcplx()), z)
    assert td(h, z)

    a1, a2, b1, b2, b3 = symbols('a1:3, b1:4')
    assert hyper((a1, a2), (b1, b2, b3), z).diff(z) == \
        a1*a2/(b1*b2*b3) * hyper((a1 + 1, a2 + 1), (b1 + 1, b2 + 1, b3 + 1), z)

    # differentiation wrt parameters is not supported
    assert hyper([z], [], z).diff(z) == Derivative(hyper([z], [], z), z)

    # hyper is unbranched wrt parameters
    assert hyper([polar_lift(z)], [polar_lift(k)], polar_lift(x)) == \
        hyper([z], [k], polar_lift(x))

    assert hyper((1, 2, 3), [3, 4], 1).is_number
    assert not hyper((1, 2, 3), [3, x], 1).is_number


def test_expand_func():
    # evaluation at 1 of Gauss' hypergeometric function:
    a1, b1, c1 = randcplx(), randcplx(), randcplx() + 5
    assert expand_func(hyper([a, b], [c], 1)) == \
        gamma(c)*gamma(-a - b + c)/(gamma(-a + c)*gamma(-b + c))
    assert abs(expand_func(hyper([a1, b1], [c1], 1))
               - hyper([a1, b1], [c1], 1)).evalf(strict=False) < 1e-10

    # hyperexpand wrapper for hyper:
    assert expand_func(hyper([], [], z)) == exp(z)
    assert expand_func(hyper([1, 2, 3], [], z)) == hyper([1, 2, 3], [], z)
    assert expand_func(meijerg([[1, 1], []], [[1], [0]], z)) == log(z + 1)
    assert expand_func(meijerg([[1, 1], []], [[], []], z)) == \
        meijerg([[1, 1], []], [[], []], z)


def replace_dummy(expr, sym):
    dum = expr.atoms(Dummy)
    if not dum:
        return expr
    assert len(dum) == 1
    return expr.xreplace({dum.pop(): sym})


def test_hyper_rewrite_sum():
    _k = Dummy('k')
    assert replace_dummy(hyper((1, 2), (1, 3), x).rewrite(Sum), _k) == \
        Sum(x**_k / factorial(_k) * RisingFactorial(2, _k) /
            RisingFactorial(3, _k), (_k, 0, oo))

    assert hyper((1, 2, 3), (-1, 3), z).rewrite(Sum) == \
        hyper((1, 2, 3), (-1, 3), z)


def test_radius_of_convergence():
    assert hyper((1, 2), [3], z).radius_of_convergence == 1
    assert hyper((1, 2), [3, 4], z).radius_of_convergence == oo
    assert hyper((1, 2, 3), [4], z).radius_of_convergence == 0
    assert hyper((0, 1, 2), [4], z).radius_of_convergence == oo
    assert hyper((-1, 1, 2), [-4], z).radius_of_convergence == 0
    assert hyper((-1, -2, 2), [-1], z).radius_of_convergence == oo
    assert hyper((-1, 2), [-1, -2], z).radius_of_convergence == 0
    assert hyper([-1, 1, 3], [-2, 2], z).radius_of_convergence == 1
    assert hyper([-1, 1], [-2, 2], z).radius_of_convergence == oo
    assert hyper([-1, 1, 3], [-2], z).radius_of_convergence == 0
    assert hyper((-1, 2, 3, 4), [], z).radius_of_convergence == oo
    assert hyper((-4, 2), [-3], z).radius_of_convergence == 0

    assert hyper([1, 1], [3], 1).convergence_statement
    assert hyper([1, 1], [2], 1).convergence_statement == false
    assert hyper([1, 1], [2], -1).convergence_statement
    assert hyper([1, 1], [1], -1).convergence_statement == false


def test_meijer():
    pytest.raises(TypeError, lambda: meijerg(1, z))
    pytest.raises(TypeError, lambda: meijerg(((1,), (2,)), (3,), (4,), z))
    pytest.raises(TypeError, lambda: meijerg((1, 2, 3), (4, 5), z))

    assert meijerg(((1, 2), (3,)), ((4,), (5,)), z) == \
        meijerg(Tuple(1, 2), Tuple(3), Tuple(4), Tuple(5), z)

    g = meijerg((1, 2), (3, 4, 5), (6, 7, 8, 9), (10, 11, 12, 13, 14), z)
    assert g.an == Tuple(1, 2)
    assert g.ap == Tuple(1, 2, 3, 4, 5)
    assert g.aother == Tuple(3, 4, 5)
    assert g.bm == Tuple(6, 7, 8, 9)
    assert g.bq == Tuple(6, 7, 8, 9, 10, 11, 12, 13, 14)
    assert g.bother == Tuple(10, 11, 12, 13, 14)
    assert g.argument == z
    assert g.nu == 75
    assert g.delta == -1
    assert g.is_commutative is True

    assert meijerg([1, 2], [3], [4], [5], z).delta == Rational(1, 2)

    # just a few checks to make sure that all arguments go where they should
    assert tn(meijerg(Tuple(), Tuple(), Tuple(0), Tuple(), -z), exp(z), z)
    assert tn(sqrt(pi)*meijerg(Tuple(), Tuple(),
                               Tuple(0), Tuple(Rational(1, 2)), z**2/4), cos(z), z)
    assert tn(meijerg(Tuple(1, 1), Tuple(), Tuple(1), Tuple(0), z),
              log(1 + z), z)

    # test exceptions
    pytest.raises(ValueError, lambda: meijerg(((3, 1), (2,)),
                                              ((oo,), (2, 0)), x))
    pytest.raises(ValueError, lambda: meijerg(((3, 1), (2,)),
                                              ((1,), (2, 0)), x))

    # differentiation
    g = meijerg((randcplx(),), (randcplx() + 2*I,), Tuple(),
                (randcplx(), randcplx()), z)
    assert td(g, z)

    g = meijerg(Tuple(), (randcplx(),), Tuple(),
                (randcplx(), randcplx()), z)
    assert td(g, z)

    g = meijerg(Tuple(), Tuple(), Tuple(randcplx()),
                Tuple(randcplx(), randcplx()), z)
    assert td(g, z)

    a1, a2, b1, b2, c1, c2, d1, d2 = symbols('a1:3, b1:3, c1:3, d1:3')
    assert meijerg((a1, a2), (b1, b2), (c1, c2), (d1, d2), z).diff(z) == \
        (meijerg((a1 - 1, a2), (b1, b2), (c1, c2), (d1, d2), z)
         + (a1 - 1)*meijerg((a1, a2), (b1, b2), (c1, c2), (d1, d2), z))/z

    assert meijerg([z, z], [], [], [], z).diff(z) == \
        Derivative(meijerg([z, z], [], [], [], z), z)

    # meijerg is unbranched wrt parameters
    assert meijerg([polar_lift(a1)], [polar_lift(a2)], [polar_lift(b1)],
                   [polar_lift(b2)], polar_lift(z)) == meijerg([a1], [a2],
                                                               [b1], [b2],
                                                               polar_lift(z))

    # integrand
    assert meijerg([a], [b], [c], [d], z).integrand(s) == \
        z**s*gamma(c - s)*gamma(-a + s + 1)/(gamma(b - s)*gamma(-d + s + 1))

    assert meijerg([[], []], [[Rational(1, 2)], [0]], 1).is_number
    assert not meijerg([[], []], [[x], [0]], 1).is_number


def test_meijerg_derivative():
    assert meijerg([], [1, 1], [0, 0, x], [], z).diff(x) == \
        log(z)*meijerg([], [1, 1], [0, 0, x], [], z) \
        + 2*meijerg([], [1, 1, 1], [0, 0, x, 0], [], z)

    y = randcplx()
    a = 5  # mpmath chokes with non-real numbers, and Mod1 with floats
    assert td(meijerg([x], [], [], [], y), x)
    assert td(meijerg([x**2], [], [], [], y), x)
    assert td(meijerg([], [x], [], [], y), x)
    assert td(meijerg([], [], [x], [], y), x)
    assert td(meijerg([], [], [], [x], y), x)
    assert td(meijerg([x], [a], [a + 1], [], y), x)
    assert td(meijerg([x], [a + 1], [a], [], y), x)
    assert td(meijerg([x, a], [], [], [a + 1], y), x)
    assert td(meijerg([x, a + 1], [], [], [a], y), x)

    b = Rational(3, 2)
    assert td(meijerg([a + 2], [b], [b - 3, x], [a], y), x)
    assert td(meijerg([x], [2, b], [1, b + 1], [], y), x)


def test_meijerg_period():
    assert meijerg([], [1], [0], [], x).get_period() == 2*pi
    assert meijerg([1], [], [], [0], x).get_period() == 2*pi
    assert meijerg([], [], [0], [], x).get_period() == 2*pi  # exp(x)
    assert meijerg(
        [], [], [0], [Rational(1, 2)], x).get_period() == 2*pi  # cos(sqrt(x))
    assert meijerg(
        [], [], [Rational(1, 2)], [0], x).get_period() == 4*pi  # sin(sqrt(x))
    assert meijerg([1, 1], [], [1], [0], x).get_period() == oo  # log(1 + x)


def test_hyper_unpolarify():
    a = exp_polar(2*pi*I)*x
    b = x
    assert hyper([], [], a).argument == b
    assert hyper([0], [], a).argument == a
    assert hyper([0], [0], a).argument == b
    assert hyper([0, 1], [0], a).argument == a


def test_hyperrep():
    # First test the base class works.
    a, b, c, d, z = symbols('a b c d z')

    class MyRep(HyperRep):
        @classmethod
        def _expr_small(cls, x):
            return a

        @classmethod
        def _expr_small_minus(cls, x):
            return b

        @classmethod
        def _expr_big(cls, x, n):
            return c*n

        @classmethod
        def _expr_big_minus(cls, x, n):
            return d*n
    assert MyRep(z).rewrite('nonrep') == Piecewise((0, abs(z) > 1), (a, True))
    assert MyRep(exp_polar(I*pi)*z).rewrite('nonrep') == \
        Piecewise((0, abs(z) > 1), (b, True))
    assert MyRep(exp_polar(2*I*pi)*z).rewrite('nonrep') == \
        Piecewise((c, abs(z) > 1), (a, True))
    assert MyRep(exp_polar(3*I*pi)*z).rewrite('nonrep') == \
        Piecewise((d, abs(z) > 1), (b, True))
    assert MyRep(exp_polar(4*I*pi)*z).rewrite('nonrep') == \
        Piecewise((2*c, abs(z) > 1), (a, True))
    assert MyRep(exp_polar(5*I*pi)*z).rewrite('nonrep') == \
        Piecewise((2*d, abs(z) > 1), (b, True))
    assert MyRep(z).rewrite('nonrepsmall') == a
    assert MyRep(exp_polar(I*pi)*z).rewrite('nonrepsmall') == b

    def t(func, hyp, z):
        """Test that func is a valid representation of hyp."""
        # First test that func agrees with hyp for small z
        if not tn(func.rewrite('nonrepsmall'), hyp, z,
                  a=Rational(-1, 2), b=Rational(-1, 2), c=Rational(1, 2), d=Rational(1, 2)):
            return False
        # Next check that the two small representations agree.
        if not tn(
            func.rewrite('nonrepsmall').subs({z: exp_polar(I*pi)*z}).replace(exp_polar, exp),
            func.subs({z: exp_polar(I*pi)*z}).rewrite('nonrepsmall'),
                z, a=Rational(-1, 2), b=Rational(-1, 2), c=Rational(1, 2), d=Rational(1, 2)):
            return False
        # Next check continuity along exp_polar(I*pi)*t
        expr = func.subs({z: exp_polar(I*pi)*z}).rewrite('nonrep')
        if abs(expr.subs({z: 1 + 1e-15}) - expr.subs({z: 1 - 1e-15})).evalf(strict=False) > 1e-10:
            return False
        # Finally check continuity of the big reps.

        def dosubs(func, a, b):
            rv = func.subs({z: exp_polar(a)*z}).rewrite('nonrep')
            return rv.subs({z: exp_polar(b)*z}).replace(exp_polar, exp)
        for n in [0, 1, 2, 3, 4, -1, -2, -3, -4]:
            expr1 = dosubs(func, 2*I*pi*n, I*pi/2)
            expr2 = dosubs(func, 2*I*pi*n + I*pi, -I*pi/2)
            if not tn(expr1, expr2, z):
                return False
            expr1 = dosubs(func, 2*I*pi*(n + 1), -I*pi/2)
            expr2 = dosubs(func, 2*I*pi*n + I*pi, I*pi/2)
            if not tn(expr1, expr2, z):
                return False
        return True

    # Now test the various representatives.
    a = Rational(1, 3)
    assert t(HyperRep_atanh(z), hyper([Rational(1, 2), 1], [Rational(3, 2)], z), z)
    assert t(HyperRep_power1(a, z), hyper([-a], [], z), z)
    assert t(HyperRep_power2(a, z), hyper([a, a - Rational(1, 2)], [2*a], z), z)
    assert t(HyperRep_log1(z), -z*hyper([1, 1], [2], z), z)
    assert t(HyperRep_asin1(z), hyper([Rational(1, 2), Rational(1, 2)], [Rational(3, 2)], z), z)
    assert t(HyperRep_asin2(z), hyper([1, 1], [Rational(3, 2)], z), z)
    assert t(HyperRep_sqrts1(a, z), hyper([-a, Rational(1, 2) - a], [Rational(1, 2)], z), z)
    assert t(HyperRep_sqrts2(a, z),
             -2*z/(2*a + 1)*hyper([-a - Rational(1, 2), -a], [Rational(1, 2)], z).diff(z), z)
    assert t(HyperRep_log2(z), -z/4*hyper([Rational(3, 2), 1, 1], [2, 2], z), z)
    assert t(HyperRep_cosasin(a, z), hyper([-a, a], [Rational(1, 2)], z), z)
    assert t(HyperRep_sinasin(a, z), 2*a*z*hyper([1 - a, 1 + a], [Rational(3, 2)], z), z)


def test_meijerg_eval():
    a = randcplx()
    arg = x*exp_polar(k*pi*I)
    expr1 = pi*meijerg([[], [(a + 1)/2]], [[a/2], [-a/2, (a + 1)/2]], arg**2/4)
    expr2 = besseli(a, arg)

    # Test that the two expressions agree for all arguments.
    for x_ in [0.5, 1.5]:
        for k_ in [0.0, 0.1, 0.3, 0.5, 0.8, 1, 5.751, 15.3]:
            assert abs((expr1 - expr2).evalf(subs={x: x_, k: k_}, strict=False)) < 1e-10
            assert abs((expr1 - expr2).evalf(subs={x: x_, k: -k_}, strict=False)) < 1e-10

    # Test continuity independently
    eps = 1e-13
    expr2 = expr1.subs({k: l})
    for x_ in [0.5, 1.5]:
        for k_ in [0.5, Rational(1, 3), 0.25, 0.75, Rational(2, 3), 1.0, 1.5]:
            assert abs((expr1 - expr2).evalf(
                       subs={x: x_, k: k_ + eps, l: k_ - eps})) < 1e-10
            assert abs((expr1 - expr2).evalf(
                       subs={x: x_, k: -k_ + eps, l: -k_ - eps})) < 1e-10

    expr = (meijerg(((0.5,), ()), ((0.5, 0, 0.5), ()), exp_polar(-I*pi)/4)
            + meijerg(((0.5,), ()), ((0.5, 0, 0.5), ()), exp_polar(I*pi)/4)) \
        / (2*sqrt(pi))
    assert (expr - pi/exp(1)).evalf(chop=True) == 0


def test_limits():
    k, x = symbols('k, x')
    assert hyper((1,), (Rational(4, 3), Rational(5, 3)), k**2).series(k) == \
        hyper((1,), (Rational(4, 3), Rational(5, 3)), 0) + \
        9*k**2*hyper((2,), (Rational(7, 3), Rational(8, 3)), 0)/20 + \
        81*k**4*hyper((3,), (Rational(10, 3), Rational(11, 3)), 0)/1120 + \
        O(k**6)  # issue sympy/sympy#6350
    assert limit(meijerg((), (), (1,), (0,), -x), x, 0) == \
        meijerg(((), ()), ((1,), (0,)), 0)  # issue sympy/sympy#6052


def test_evalf():
    assert hyper((-1, 1), (-1,), 1).evalf() == Float('2.0')

    e = meijerg([], [], [], [], (exp_polar(-I*pi))*cos(exp_polar(-I*pi)))
    assert e.evalf() == e



from diofant import (Derivative, E, I, Integer, O, cos, exp, gamma, log, pi,
                     sin, tanh)
from diofant.abc import x, y


__all__ = ()


def test_sin():
    e = sin(x).series(x, n=None)
    assert next(e) == x
    assert next(e) == -x**3/6
    assert next(e) == x**5/120


def test_cos():
    e = cos(x).series(x, n=None)
    assert next(e) == 1
    assert next(e) == -x**2/2
    assert next(e) == x**4/24


def test_exp():
    e = exp(x).series(x, n=None)
    assert next(e) == 1
    assert next(e) == x
    assert next(e) == x**2/2
    assert next(e) == x**3/6


def test_exp2():
    e = exp(cos(x)).series(x, n=None)
    assert next(e) == E
    assert next(e) == -E*x**2/2
    assert next(e) == E*x**4/6
    assert next(e) == -31*E*x**6/720


def test_simple():
    assert list(x.series(n=None)) == [x]
    assert list(Integer(1).series(x, n=None)) == [1]
    assert not next((x/(x + y)).series(y, n=None)).has(O)
    assert [t.doit() for t in Derivative(1 + x, x).series(x, n=None)] == [0, 1]

    # issue sympy/sympy#5183
    s = (x + 1/x).series(n=None)
    assert list(s) == [1/x, x]

    # issue sympy/sympy#5183
    s = (x + 1/x).series(n=None)
    assert next((x + x**2).series(n=None)) == x
    assert next(((1 + x)**7).series(x, n=None)) == 1
    assert next((sin(x + y)).series(x, n=3).series(y, n=None)) == x
    # it would be nice if all terms were grouped, but in the
    # following case that would mean that all the terms would have
    # to be known since, for example, every term has a constant in it.
    s = ((1 + x)**7).series(x, 1, n=None)
    assert [next(s) for i in range(2)] == [128, -448 + 448*x]


def test_tanh():
    # issue sympy/sympy#6999
    s = tanh(x).series(x, x0=1, n=None)
    assert next(s) == tanh(1)
    assert next(s) == x - (x - 1)*tanh(1)**2 - 1
    assert next(s) == -(x - 1)**2*tanh(1) + (x - 1)**2*tanh(1)**3
    assert next(s) == -(x - 1)**3*tanh(1)**4 - (x - 1)**3/3 + \
        4*(x - 1)**3*tanh(1)**2/3


def test_sympyissue_21859():
    e = -exp(-x*log(y)/2)*exp(-5*I*pi*x/6)*gamma(5*x/6 + 1)/(gamma(x/3 + 1)*gamma(x/2 + 1))
    s = e.series(x, n=None)

    assert next(s) == -1
    assert next(s) == x*log(y)/2 + 5*I*pi*x/6



"""
Finite difference weights
=========================

This module implements an algorithm for efficient generation of finite
difference weights for ordinary differentials of functions for
derivatives from 0 (interpolation) up to arbitrary order.

The core algorithm is provided in the finite difference weight generating
function (finite_diff_weights), and two convenience functions are provided
for:

- estimating a derivative (or interpolate) directly from a series of points
    is also provided (``apply_finite_diff``).
- making a finite difference approximation of a Derivative instance
    (``as_finite_diff``).

"""

from ..core import Integer
from ..core.compatibility import iterable


def finite_diff_weights(order, x_list, x0=Integer(0)):
    """
    Calculates the finite difference weights for an arbitrarily
    spaced one-dimensional grid (x_list) for derivatives at 'x0'
    of order 0, 1, ..., up to 'order' using a recursive formula.
    Order of accuracy is at least len(x_list) - order, if x_list
    is defined accurately.


    Parameters
    ==========

    order: int
        Up to what derivative order weights should be calculated.
        0 corresponds to interpolation.
    x_list: sequence
        Sequence of (unique) values for the independent variable.
        It is useful (but not necessary) to order x_list from
        nearest to furthest from x0; see examples below.
    x0: Number or Symbol
        Root or value of the independent variable for which the finite
        difference weights should be generated. Defaults to Integer(0).

    Returns
    =======

    list
        A list of sublists, each corresponding to coefficients for
        increasing derivative order, and each containing lists of
        coefficients for increasing subsets of x_list.

    Examples
    ========

    >>> res = finite_diff_weights(1, [-Rational(1, 2), Rational(1, 2),
    ...                               Rational(3, 2), Rational(5, 2)], 0)
    >>> res
    [[[1, 0, 0, 0],
      [1/2, 1/2, 0, 0],
      [3/8, 3/4, -1/8, 0],
      [5/16, 15/16, -5/16, 1/16]],
     [[0, 0, 0, 0],
      [-1, 1, 0, 0],
      [-1, 1, 0, 0],
      [-23/24, 7/8, 1/8, -1/24]]]
    >>> res[0][-1]  # FD weights for 0th derivative, using full x_list
    [5/16, 15/16, -5/16, 1/16]
    >>> res[1][-1]  # FD weights for 1st derivative
    [-23/24, 7/8, 1/8, -1/24]
    >>> res[1][-2]  # FD weights for 1st derivative, using x_list[:-1]
    [-1, 1, 0, 0]
    >>> res[1][-1][0]  # FD weight for 1st deriv. for x_list[0]
    -23/24
    >>> res[1][-1][1]  # FD weight for 1st deriv. for x_list[1], etc.
    7/8

    Each sublist contains the most accurate formula at the end.
    Note, that in the above example res[1][1] is the same as res[1][2].
    Since res[1][2] has an order of accuracy of
    len(x_list[:3]) - order = 3 - 1 = 2, the same is true for res[1][1]!

    >>> res = finite_diff_weights(1, [Integer(0), Integer(1), -Integer(1),
    ...                               Integer(2), -Integer(2)], 0)[1]
    >>> res
    [[0, 0, 0, 0, 0],
     [-1, 1, 0, 0, 0],
     [0, 1/2, -1/2, 0, 0],
     [-1/2, 1, -1/3, -1/6, 0],
     [0, 2/3, -2/3, -1/12, 1/12]]
    >>> res[0]  # no approximation possible, using x_list[0] only
    [0, 0, 0, 0, 0]
    >>> res[1]  # classic forward step approximation
    [-1, 1, 0, 0, 0]
    >>> res[2]  # classic centered approximation
    [0, 1/2, -1/2, 0, 0]
    >>> res[3:]  # higher order approximations
    [[-1/2, 1, -1/3, -1/6, 0], [0, 2/3, -2/3, -1/12, 1/12]]

    Let us compare this to a differently defined x_list. Pay attention to
    foo[i][k] corresponding to the gridpoint defined by x_list[k].

    >>> foo = finite_diff_weights(1, [-Integer(2), -Integer(1), Integer(0),
    ...                               Integer(1), Integer(2)], 0)[1]
    >>> foo
    [[0, 0, 0, 0, 0],
     [-1, 1, 0, 0, 0],
     [1/2, -2, 3/2, 0, 0],
     [1/6, -1, 1/2, 1/3, 0],
     [1/12, -2/3, 0, 2/3, -1/12]]
    >>> foo[1]  # not the same and of lower accuracy as res[1]!
    [-1, 1, 0, 0, 0]
    >>> foo[2]  # classic double backward step approximation
    [1/2, -2, 3/2, 0, 0]
    >>> foo[4]  # the same as res[4]
    [1/12, -2/3, 0, 2/3, -1/12]

    Note that, unless you plan on using approximations based on subsets of
    x_list, the order of gridpoints does not matter.


    The capability to generate weights at arbitrary points can be
    used e.g. to minimize Runge's phenomenon by using Chebyshev nodes:

    >>> N, h = 4, symbols('h')
    >>> x_list = [x + h*cos(i*pi/(N)) for i in range(N, -1, -1)]  # chebyshev nodes
    >>> x_list
    [-h + x, -sqrt(2)*h/2 + x, x, sqrt(2)*h/2 + x, h + x]
    >>> mycoeffs = finite_diff_weights(1, x_list, 0)[1][4]
    >>> [simplify(c) for c in mycoeffs]
    [(h**3/2 + h**2*x - 3*h*x**2 - 4*x**3)/h**4,
    (-sqrt(2)*h**3 - 4*h**2*x + 3*sqrt(2)*h*x**2 + 8*x**3)/h**4,
    6*x/h**2 - 8*x**3/h**4,
    (sqrt(2)*h**3 - 4*h**2*x - 3*sqrt(2)*h*x**2 + 8*x**3)/h**4,
    (-h**3/2 + h**2*x + 3*h*x**2 - 4*x**3)/h**4]

    Notes
    =====

    If weights for a finite difference approximation of 3rd order
    derivative is wanted, weights for 0th, 1st and 2nd order are
    calculated "for free", so are formulae using subsets of x_list.
    This is something one can take advantage of to save computational cost.
    Be aware that one should define x_list from nearest to farest from
    x_list. If not, subsets of x_list will yield poorer approximations,
    which might not grand an order of accuracy of len(x_list) - order.

    See also
    ========

    diofant.calculus.finite_diff.apply_finite_diff


    References
    ==========

    * Generation of Finite Difference Formulas on Arbitrarily Spaced
      Grids, Bengt Fornberg; Mathematics of computation; 51; 184;
      (1988); 699-706; doi:10.1090/S0025-5718-1988-0935077-0

    """
    # The notation below closely corresponds to the one used in the paper.
    if order < 0:
        raise ValueError('Negative derivative order illegal.')
    if int(order) != order:
        raise ValueError('Non-integer order illegal')
    M = order
    N = len(x_list) - 1
    delta = [[[0 for nu in range(N+1)] for n in range(N+1)] for
             m in range(M+1)]
    delta[0][0][0] = Integer(1)
    c1 = Integer(1)
    for n in range(1, N+1):
        c2 = Integer(1)
        for nu in range(n):
            c3 = x_list[n]-x_list[nu]
            c2 = c2 * c3
            if n <= M:
                delta[n][n-1][nu] = 0
            for m in range(min(n, M)+1):
                delta[m][n][nu] = (x_list[n]-x0)*delta[m][n-1][nu] -\
                    m*delta[m-1][n-1][nu]
                delta[m][n][nu] /= c3
        for m in range(min(n, M)+1):
            delta[m][n][n] = c1/c2*(m*delta[m-1][n-1][n-1] -
                                    (x_list[n-1]-x0)*delta[m][n-1][n-1])
        c1 = c2
    return delta


def apply_finite_diff(order, x_list, y_list, x0=Integer(0)):
    r"""
    Calculates the finite difference approximation of
    the derivative of requested order at x0 from points
    provided in x_list and y_list.

    Parameters
    ==========

    order: int
        order of derivative to approximate. 0 corresponds to interpolation.
    x_list: sequence
        Sequence of (unique) values for the independent variable.
    y_list: sequence
        The function value at corresponding values for the independent
        variable in x_list.
    x0: Number or Symbol
        At what value of the independent variable the derivative should be
        evaluated. Defaults to Integer(0).

    Returns
    =======

    diofant.core.add.Add or diofant.core.numbers.Number
        The finite difference expression approximating the requested
        derivative order at x0.

    Examples
    ========

    >>> def cube(arg):
    ...     return (1.0*arg)**3
    >>> xlist = range(-3, 4)
    >>> apply_finite_diff(2, xlist, list(map(cube, xlist)), 2) - 12
    -3.55271367880050e-15

    we see that the example above only contain rounding errors.
    apply_finite_diff can also be used on more abstract objects:

    >>> x, y = map(IndexedBase, 'xy')
    >>> i = Idx('i')
    >>> x_list, y_list = zip(*[(x[i + j], y[i + j]) for j in range(-1, 2)])
    >>> apply_finite_diff(1, x_list, y_list, x[i])
    (-1 + (x[i + 1] - x[i])/(-x[i - 1] + x[i]))*y[i]/(x[i + 1] - x[i]) +
    (-x[i - 1] + x[i])*y[i + 1]/((-x[i - 1] + x[i + 1])*(x[i + 1] - x[i])) -
    (x[i + 1] - x[i])*y[i - 1]/((-x[i - 1] + x[i + 1])*(-x[i - 1] + x[i]))

    Notes
    =====

    Order = 0 corresponds to interpolation.
    Only supply so many points you think makes sense
    to around x0 when extracting the derivative (the function
    need to be well behaved within that region). Also beware
    of Runge's phenomenon.

    See also
    ========

    diofant.calculus.finite_diff.finite_diff_weights

    References
    ==========

    Fortran 90 implementation with Python interface for numerics: finitediff_

    .. _finitediff: https://github.com/bjodah/finitediff

    """
    # In the original paper the following holds for the notation:
    # M = order
    # N = len(x_list) - 1

    N = len(x_list) - 1
    if len(x_list) != len(y_list):
        raise ValueError('x_list and y_list not equal in length.')

    delta = finite_diff_weights(order, x_list, x0)

    derivative = 0
    for nu in range(len(x_list)):
        derivative += delta[order][N][nu]*y_list[nu]
    return derivative


def as_finite_diff(derivative, points=1, x0=None, wrt=None):
    r"""
    Returns an approximation of a derivative of a function in
    the form of a finite difference formula. The expression is a
    weighted sum of the function at a number of discrete values of
    (one of) the independent variable(s).

    Parameters
    ==========

    derivative: a Derivative instance (needs to have an variables
        and expr attribute).

    points: sequence or coefficient, optional
        If sequence: discrete values (length >= order+1) of the
        independent variable used for generating the finite
        difference weights.
        If it is a coefficient, it will be used as the step-size
        for generating an equidistant sequence of length order+1
        centered around x0. default: 1 (step-size 1)

    x0: number or Symbol, optional
        the value of the independent variable (wrt) at which the
        derivative is to be approximated. default: same as wrt

    wrt: Symbol, optional
        "with respect to" the variable for which the (partial)
        derivative is to be approximated for. If not provided it
        is required that the Derivative is ordinary. default: None

    Examples
    ========

    >>> h = symbols('h')
    >>> as_finite_diff(f(x).diff(x))
    -f(x - 1/2) + f(x + 1/2)

    The default step size and number of points are 1 and ``order + 1``
    respectively. We can change the step size by passing a symbol
    as a parameter:

    >>> as_finite_diff(f(x).diff(x), h)
    -f(-h/2 + x)/h + f(h/2 + x)/h

    We can also specify the discretized values to be used in a sequence:

    >>> as_finite_diff(f(x).diff(x), [x, x+h, x+2*h])
    -3*f(x)/(2*h) + 2*f(h + x)/h - f(2*h + x)/(2*h)

    The algorithm is not restricted to use equidistant spacing, nor
    do we need to make the approximation around x0, but we can get
    an expression estimating the derivative at an offset:

    >>> e, sq2 = exp(1), sqrt(2)
    >>> xl = [x-h, x+h, x+e*h]
    >>> as_finite_diff(f(x).diff((x, 1)), xl, x+h*sq2)
    2*h*f(E*h + x)*((h + sqrt(2)*h)/(2*h) -
    (-sqrt(2)*h + h)/(2*h))/((-h + E*h)*(h + E*h)) +
    f(-h + x)*(-(-sqrt(2)*h + h)/(2*h) - (-sqrt(2)*h + E*h)/(2*h))/(h +
    E*h) + f(h + x)*(-(h + sqrt(2)*h)/(2*h) + (-sqrt(2)*h +
    E*h)/(2*h))/(-h + E*h)

    Partial derivatives are also supported:

    >>> d2fdxdy = f(x, y).diff(x, y)
    >>> as_finite_diff(d2fdxdy, wrt=x)
    -f(x - 1/2, y) + f(x + 1/2, y)

    See also
    ========

    diofant.calculus.finite_diff.apply_finite_diff
    diofant.calculus.finite_diff.finite_diff_weights

    """
    if wrt is None:
        wrt = derivative.variables[0]
        # we need Derivative to be univariate to guess wrt
        if any(v != wrt for v in derivative.variables):
            raise ValueError('if the function is not univariate' +
                             ' then `wrt` must be given')

    order = derivative.variables.count(wrt)

    if x0 is None:
        x0 = wrt

    if not iterable(points):
        # points is simply the step-size, let's make it a
        # equidistant sequence centered around x0
        if order % 2 == 0:
            # even order => odd number of points, grid point included
            points = [x0 + points*i for i
                      in range(-order//2, order//2 + 1)]
        else:
            # odd order => even number of points, half-way wrt grid point
            points = [x0 + points*i/Integer(2) for i
                      in range(-order, order + 1, 2)]

    if len(points) < order+1:
        raise ValueError(f'Too few points for order {order:d}')
    return apply_finite_diff(order, points, [
        derivative.expr.subs({wrt: x}) for x in points], x0)



"""Implementation of :class:`RealField` class."""

from __future__ import annotations

import mpmath

from ..core import Float
from ..polys.polyerrors import CoercionFailed
from .characteristiczero import CharacteristicZero
from .field import Field
from .mpelements import MPContext
from .simpledomain import SimpleDomain


class RealField(CharacteristicZero, SimpleDomain, Field):
    """Real numbers up to the given precision."""

    rep = 'RR'

    is_RealField = True

    is_Exact = False
    is_Numerical = True

    _default_precision = 53

    @property
    def has_default_precision(self):
        return self.precision == self._default_precision

    @property
    def precision(self):
        return self._context.prec

    @property
    def dps(self):
        return self._context.dps

    @property
    def tolerance(self):
        return self._context.tolerance

    def __new__(cls, prec=_default_precision, dps=None, tol=None):
        context = MPContext(prec, dps, tol)

        obj = super().__new__(cls)

        try:
            obj.dtype = _reals_cache[(context.prec, context.tolerance)]
        except KeyError:
            _reals_cache[(context.prec, context.tolerance)] = obj.dtype = context.mpf

        context._parent = obj
        obj._context = context
        obj._hash = hash((cls.__name__, obj.dtype, context.prec, context.tolerance))

        obj.zero = obj.dtype(0)
        obj.one = obj.dtype(1)

        return obj

    def __getnewargs_ex__(self):
        return (), {'prec': self.precision,
                    'tol': mpmath.mpf(self.tolerance._mpf_)}

    def __eq__(self, other):
        return (isinstance(other, RealField)
                and self.precision == other.precision
                and self.tolerance == other.tolerance)

    def __hash__(self):
        return self._hash

    def to_expr(self, element):
        return Float(element, self.dps)

    def from_expr(self, expr):
        number = expr.evalf(self.dps)

        if number.is_Number:
            return self.dtype(number)
        else:
            raise CoercionFailed(f'expected real number, got {expr}')

    def _from_PythonIntegerRing(self, element, base):
        return self.dtype(element)
    _from_GMPYIntegerRing = _from_PythonIntegerRing

    def _from_PythonRationalField(self, element, base):
        return self.dtype(element.numerator) / element.denominator
    _from_GMPYRationalField = _from_PythonRationalField

    def _from_AlgebraicField(self, element, base):
        return self.from_expr(base.to_expr(element))

    def _from_RealField(self, element, base):
        if self == base:
            return element
        else:
            return self.dtype(element)

    def _from_ComplexField(self, element, base):
        if not element.imag:
            return self.dtype(element.real)

    def to_rational(self, element, limit=True):
        """Convert a real number to rational number."""
        return self._context.to_rational(element, limit)

    def get_exact(self):
        from . import QQ
        return QQ

    def gcd(self, a, b):
        return self.one

    def almosteq(self, a, b, tolerance=None):
        """Check if ``a`` and ``b`` are almost equal."""
        return self._context.almosteq(a, b, tolerance)


_reals_cache: dict[tuple, RealField] = {}


RR = RealField()



"""Options manager for :class:`~diofant.polys.polytools.Poly` and public API functions."""

from __future__ import annotations

import re

from ..core import Basic, I
from ..core.sympify import sympify
from ..utilities import has_dups, numbered_symbols, topological_sort
from .polyerrors import FlagError, GeneratorsError, OptionError


__all__ = 'Options', 'Order'


class Option:
    """Base class for all kinds of options."""

    option: str

    is_Flag = False

    requires: list[str] = []
    excludes: list[str] = []

    after: list[str] = []
    before: list[str] = []

    @classmethod
    def default(cls):
        return

    @classmethod
    def preprocess(cls, option):
        return   # pragma: no cover

    @classmethod
    def postprocess(cls, options):
        return


class Flag(Option):
    """Base class for all kinds of flags."""

    is_Flag = True


class BooleanOption(Option):
    """An option that must have a boolean value or equivalent assigned."""

    @classmethod
    def preprocess(cls, option):
        if option in [True, False]:
            return bool(option)
        else:
            raise OptionError(f"'{cls.option}' must have a boolean value "
                              f'assigned, got {option}')


class OptionType(type):
    """Base type for all options that does registers options."""

    def __init__(cls, *args, **kwargs):
        super().__init__(cls)

        @property
        def getter(a):
            try:
                return a[cls.option]
            except KeyError:
                return cls.default()

        setattr(Options, cls.option, getter)
        Options.__options__[cls.option] = cls


class Options(dict):
    """
    Options manager for polynomial manipulation module.

    Examples
    ========

    >>> Options((x, y, z), {'domain': 'ZZ'})
    {'auto': False, 'domain': ZZ, 'gens': (x, y, z)}

    >>> build_options((x, y, z), {'domain': 'ZZ'})
    {'auto': False, 'domain': ZZ, 'gens': (x, y, z)}

    **Options**

    * Expand --- boolean option
    * Gens --- option
    * Wrt --- option
    * Sort --- option
    * Order --- option
    * Field --- boolean option
    * Greedy --- boolean option
    * Domain --- option
    * Split --- boolean option
    * Gaussian --- boolean option
    * Extension --- option
    * Modulus --- option
    * Symmetric --- boolean option
    * Strict --- boolean option

    **Flags**

    * Auto --- boolean flag
    * Frac --- boolean flag
    * Formal --- boolean flag
    * Polys --- boolean flag
    * Include --- boolean flag
    * All --- boolean flag
    * Gen --- flag

    """

    __order__: list[str] | None = None
    __options__: dict[str, type[Option]] = {}

    def __init__(self, gens, args, flags=None, strict=False):
        dict.__init__(self)

        if gens and args.get('gens', ()):
            raise OptionError("both '*gens' and keyword "
                              "argument 'gens' supplied")
        if gens:
            args = dict(args)
            args['gens'] = gens

        defaults = args.pop('defaults', {})

        def preprocess_options(args):
            for option, value in args.items():
                try:
                    cls = self.__options__[option]
                except KeyError as exc:
                    raise OptionError(f"'{option}' is not a "
                                      'valid option') from exc

                if issubclass(cls, Flag):
                    if strict and (flags is None or option not in flags):
                        raise OptionError(f"'{option}' flag is not "
                                          'allowed in this context')

                if value is not None:
                    self[option] = cls.preprocess(value)

        preprocess_options(args)

        for key in dict(defaults):
            if key in self:
                del defaults[key]
            else:
                for option in self:
                    cls = self.__options__[option]

                    if key in cls.excludes:
                        del defaults[key]
                        break

        preprocess_options(defaults)

        for option in self:
            cls = self.__options__[option]

            for exclude_option in cls.excludes:
                if self.get(exclude_option) is not None:
                    raise OptionError(f"'{option}' option is not allowed together with '{exclude_option}'")

        for option in self.__order__:  # pylint: disable=not-an-iterable
            self.__options__[option].postprocess(self)

    @classmethod
    def _init_dependencies_order(cls):
        """Resolve the order of options' processing."""
        if cls.__order__ is None:
            vertices, edges = [], set()

            for name, option in cls.__options__.items():
                vertices.append(name)

                for _name in option.after:
                    edges.add((_name, name))

                for _name in option.before:
                    edges.add((name, _name))

            try:
                cls.__order__ = topological_sort((vertices, list(edges)))
            except ValueError as exc:
                raise RuntimeError('cycle detected in diofant.polys'
                                   ' options framework') from exc

    def clone(self, updates={}):
        """Clone ``self`` and update specified options."""
        obj = dict.__new__(self.__class__)

        for option, value in self.items():
            obj[option] = value

        for option, value in updates.items():
            obj[option] = value

        return obj

    def __setattr__(self, attr, value):
        if attr in self.__options__:
            self[attr] = value
        else:
            super().__setattr__(attr, value)

    @property
    def args(self):
        args = {}

        for option, value in self.items():
            if value is not None and option != 'gens':
                cls = self.__options__[option]

                if not issubclass(cls, Flag):
                    args[option] = value

        return args

    @property
    def options(self):
        options = {}
        for option, cls in self.__options__.items():
            if not issubclass(cls, Flag):
                options[option] = getattr(self, option)
        return options

    @property
    def flags(self):
        flags = {}

        for option, cls in self.__options__.items():
            if issubclass(cls, Flag):
                flags[option] = getattr(self, option)

        return flags


class Expand(BooleanOption, metaclass=OptionType):
    """``expand`` option to polynomial manipulation functions."""

    option = 'expand'

    @classmethod
    def default(cls):
        return True


class Gens(Option, metaclass=OptionType):
    """``gens`` option to polynomial manipulation functions."""

    option = 'gens'

    @classmethod
    def default(cls):
        return ()

    @classmethod
    def preprocess(cls, option):
        if isinstance(option, Basic):
            option = option,

        if option == (None,):
            return ()
        elif has_dups(option):
            raise GeneratorsError(f'duplicated generators: {option}')
        elif any(gen.is_commutative is False for gen in option):
            raise GeneratorsError(f'non-commutative generators: {option}')
        else:
            return tuple(option)


class Wrt(Option, metaclass=OptionType):
    """``wrt`` option to polynomial manipulation functions."""

    option = 'wrt'

    _re_split = re.compile(r'\s*,\s*|\s+')

    @classmethod
    def preprocess(cls, option):
        if isinstance(option, Basic):
            return [str(option)]
        elif isinstance(option, str):
            option = option.strip()
            if option.endswith(','):
                raise OptionError('Bad input: missing parameter.')
            if not option:
                return []
            return list(cls._re_split.split(option))
        elif hasattr(option, '__getitem__'):
            return list(map(str, option))
        else:
            raise OptionError("invalid argument for 'wrt' option")


class Sort(Option, metaclass=OptionType):
    """``sort`` option to polynomial manipulation functions."""

    option = 'sort'

    @classmethod
    def default(cls):
        return []

    @classmethod
    def preprocess(cls, option):
        if isinstance(option, str):
            return [gen.strip() for gen in option.split('>')]
        elif hasattr(option, '__getitem__'):
            return list(map(str, option))
        else:
            raise OptionError("invalid argument for 'sort' option")


class Order(Option, metaclass=OptionType):
    """``order`` option to polynomial manipulation functions."""

    option = 'order'

    @classmethod
    def default(cls):
        from .orderings import lex
        return lex

    @classmethod
    def preprocess(cls, option):
        from .orderings import monomial_key
        return monomial_key(option)


class Field(BooleanOption, metaclass=OptionType):
    """``field`` option to polynomial manipulation functions."""

    option = 'field'

    excludes = ['domain', 'split', 'gaussian']


class Greedy(BooleanOption, metaclass=OptionType):
    """``greedy`` option to polynomial manipulation functions."""

    option = 'greedy'

    excludes = ['domain', 'split', 'gaussian', 'extension', 'modulus']


class Composite(BooleanOption, metaclass=OptionType):
    """``composite`` option to polynomial manipulation functions."""

    option = 'composite'

    @classmethod
    def default(cls):
        return

    excludes = ['domain', 'split', 'gaussian', 'modulus']


class Domain(Option, metaclass=OptionType):
    """``domain`` option to polynomial manipulation functions."""

    option = 'domain'

    excludes = ['field', 'greedy', 'split', 'gaussian', 'extension']

    after = ['gens']

    _re_realfield = re.compile(r'^(R|RR)(_(\d+))?$')
    _re_complexfield = re.compile(r'^(C|CC)(_(\d+))?$')
    _re_finitefield = re.compile(r'^(FF|GF)\((\d+)\)$')
    _re_polynomial = re.compile(r'^(Z|ZZ|Q|QQ)\[(.+)\]$')
    _re_fraction = re.compile(r'^(Z|ZZ|Q|QQ)\((.+)\)$')
    _re_algebraic = re.compile(r'^(Q|QQ)\<(.+)\>$')

    @classmethod
    def preprocess(cls, option):
        from .. import domains
        if isinstance(option, domains.Domain):
            return option
        elif isinstance(option, str):
            if option in ['Z', 'ZZ']:
                return domains.ZZ

            if option in ['Q', 'QQ']:
                return domains.QQ

            if option == 'EX':
                return domains.EX

            r = cls._re_realfield.match(option)

            if r is not None:
                _, _, prec = r.groups()

                if prec is None:
                    return domains.RR
                else:
                    return domains.RealField(int(prec))

            r = cls._re_complexfield.match(option)

            if r is not None:
                _, _, prec = r.groups()

                if prec is None:
                    return domains.CC
                else:
                    return domains.ComplexField(int(prec))

            r = cls._re_finitefield.match(option)

            if r is not None:
                return domains.FF(int(r.groups()[1]))

            r = cls._re_polynomial.match(option)

            if r is not None:
                ground, gens = r.groups()

                gens = list(map(sympify, gens.split(',')))

                if ground in ['Z', 'ZZ']:
                    return domains.ZZ.inject(*gens)
                else:
                    return domains.QQ.inject(*gens)

            r = cls._re_fraction.match(option)

            if r is not None:
                ground, gens = r.groups()

                gens = list(map(sympify, gens.split(',')))

                if ground in ['Z', 'ZZ']:
                    return domains.ZZ.inject(*gens).field
                else:
                    return domains.QQ.inject(*gens).field

            r = cls._re_algebraic.match(option)

            if r is not None:
                gens = list(map(sympify, r.groups()[1].split(',')))
                return domains.QQ.algebraic_field(*gens)

        raise OptionError('expected a valid domain specification, '
                          f'got {option}')

    @classmethod
    def postprocess(cls, options):
        from .. import domains
        from ..domains.compositedomain import CompositeDomain

        if 'gens' in options and 'domain' in options and isinstance(options['domain'], CompositeDomain) and \
                (set(options['domain'].symbols) & set(options['gens'])):
            raise GeneratorsError('ground domain and generators '
                                  'interfere together')
        if ('gens' not in options or not options['gens']) and \
                'domain' in options and options['domain'] == domains.EX:
            raise GeneratorsError('you have to provide generators because'
                                  ' EX domain was requested')


class Split(BooleanOption, metaclass=OptionType):
    """``split`` option to polynomial manipulation functions."""

    option = 'split'

    excludes = ['field', 'greedy', 'domain', 'gaussian', 'extension', 'modulus']

    @classmethod
    def postprocess(cls, options):
        if 'split' in options:
            raise NotImplementedError("'split' option is not implemented yet")


class Gaussian(BooleanOption, metaclass=OptionType):
    """``gaussian`` option to polynomial manipulation functions."""

    option = 'gaussian'

    excludes = ['field', 'greedy', 'domain', 'split', 'extension', 'modulus']

    @classmethod
    def postprocess(cls, options):
        if 'gaussian' in options and options['gaussian'] is True:
            options['extension'] = {I}
            Extension.postprocess(options)


class Extension(Option, metaclass=OptionType):
    """``extension`` option to polynomial manipulation functions."""

    option = 'extension'

    excludes = ['greedy', 'domain', 'split', 'gaussian', 'modulus']

    @classmethod
    def preprocess(cls, option):
        if option == 1:
            return bool(option)
        elif option == 0:
            return bool(option)
        else:
            if not hasattr(option, '__iter__'):
                option = {option}
            else:
                if not option:
                    option = None
                else:
                    option = set(option)

            return option

    @classmethod
    def postprocess(cls, options):
        from .. import domains
        if 'extension' in options and options['extension'] not in (True, False):
            options['domain'] = domains.QQ.algebraic_field(
                *options['extension'])


class Modulus(Option, metaclass=OptionType):
    """``modulus`` option to polynomial manipulation functions."""

    option = 'modulus'

    excludes = ['greedy', 'split', 'domain', 'gaussian', 'extension']

    @classmethod
    def preprocess(cls, option):
        option = sympify(option)

        if option.is_Integer and option > 0:
            return int(option)
        else:
            raise OptionError(
                f"'modulus' must a positive integer, got {option}")

    @classmethod
    def postprocess(cls, options):
        from .. import domains
        if 'modulus' in options:
            modulus = options['modulus']
            options['domain'] = domains.FF(modulus)


class Strict(BooleanOption, metaclass=OptionType):
    """``strict`` option to polynomial manipulation functions."""

    option = 'strict'

    @classmethod
    def default(cls):
        return True


class Auto(BooleanOption, Flag, metaclass=OptionType):
    """``auto`` flag to polynomial manipulation functions."""

    option = 'auto'

    after = ['field', 'domain', 'extension', 'gaussian']

    @classmethod
    def default(cls):
        return True

    @classmethod
    def postprocess(cls, options):
        if ('domain' in options or 'field' in options) and 'auto' not in options:
            options['auto'] = False


class Frac(BooleanOption, Flag, metaclass=OptionType):
    """``frac`` option to polynomial manipulation functions."""

    option = 'frac'

    @classmethod
    def default(cls):
        return False


class Formal(BooleanOption, Flag, metaclass=OptionType):
    """``formal`` flag to polynomial manipulation functions."""

    option = 'formal'

    @classmethod
    def default(cls):
        return False


class Polys(BooleanOption, Flag, metaclass=OptionType):
    """``polys`` flag to polynomial manipulation functions."""

    option = 'polys'


class Include(BooleanOption, Flag, metaclass=OptionType):
    """``include`` flag to polynomial manipulation functions."""

    option = 'include'

    @classmethod
    def default(cls):
        return False


class All(BooleanOption, Flag, metaclass=OptionType):
    """``all`` flag to polynomial manipulation functions."""

    option = 'all'

    @classmethod
    def default(cls):
        return False


class Gen(Flag, metaclass=OptionType):
    """``gen`` flag to polynomial manipulation functions."""

    option = 'gen'

    @classmethod
    def default(cls):
        return 0

    @classmethod
    def preprocess(cls, option):
        if isinstance(option, (Basic, int)):
            return option
        else:
            raise OptionError("invalid argument for 'gen' option")


class Symbols(Flag, metaclass=OptionType):
    """``symbols`` flag to polynomial manipulation functions."""

    option = 'symbols'

    @classmethod
    def default(cls):
        return numbered_symbols('s', start=1)

    @classmethod
    def preprocess(cls, option):
        if hasattr(option, '__iter__'):
            return iter(option)
        else:
            raise OptionError('expected an iterator or '
                              f'iterable container, got {option}')


class Method(Flag, metaclass=OptionType):
    """``method`` flag to polynomial manipulation functions."""

    option = 'method'

    @classmethod
    def preprocess(cls, option):
        if isinstance(option, str):
            return option.lower()
        else:
            raise OptionError(f'expected a string, got {option}')


def build_options(gens, args=None):
    """Construct options from keyword arguments or ... options."""
    if args is None:
        gens, args = (), gens

    if len(args) != 1 or 'opt' not in args or gens:
        return Options(gens, args)
    else:
        return args['opt']


def allowed_flags(args, flags):
    """
    Allow specified flags to be used in the given context.

    Examples
    ========

    >>> allowed_flags({'domain': ZZ}, [])

    >>> allowed_flags({'domain': ZZ, 'frac': True}, [])
    Traceback (most recent call last):
    ...
    FlagError: 'frac' flag is not allowed in this context

    >>> allowed_flags({'domain': ZZ, 'frac': True}, ['frac'])

    """
    flags = set(flags)

    for arg in args:
        try:
            if Options.__options__[arg].is_Flag and arg not in flags:
                raise FlagError(f"'{arg}' flag is not allowed "
                                'in this context')
        except KeyError as exc:
            raise OptionError(f"'{arg}' is not a valid option") from exc


def set_defaults(options, **defaults):
    """Update options with default values."""
    if 'defaults' not in options:
        options = dict(options)
        options['defaults'] = defaults

    return options


Options._init_dependencies_order()



"""
Python shell for Diofant.

This is just a normal Python shell (IPython shell if you have the
IPython package installed), that adds default imports and run
some initialization code.
"""

import argparse
import ast
import atexit
import code
import os
import readline
import rlcompleter

from diofant.interactive.session import (AutomaticSymbols,
                                         IntegerDivisionWrapper)


__all__ = ()


parser = argparse.ArgumentParser(description=__doc__,
                                 prog='python -m diofant')
parser.add_argument('--no-wrap-division',
                    help="Don't wrap integer divisions with Fraction",
                    action='store_true')
parser.add_argument('-a', '--auto-symbols',
                    help="Automatically create missing Symbol's",
                    action='store_true')
parser.add_argument('--no-ipython', help="Don't use IPython",
                    action='store_true')


def main():
    args, ipython_args = parser.parse_known_args()

    lines = ['from diofant import *',
             'init_printing()',
             "a, b, c, d, t, x, y, z = symbols('a:d t x:z')",
             "k, m, n = symbols('k m n', integer=True)",
             "f, g, h = symbols('f g h', cls=Function)",
             'init_printing(pretty_print=True, use_unicode=True)']

    try:
        import IPython
        import traitlets
    except ImportError:
        args.no_ipython = True

    if not args.no_ipython:
        config = traitlets.config.loader.Config()
        shell = config.InteractiveShell
        ast_transformers = shell.ast_transformers
        if not args.no_wrap_division:
            ast_transformers.append(IntegerDivisionWrapper())
        shell.confirm_exit = False
        config.TerminalIPythonApp.display_banner = False
        config.TerminalInteractiveShell.autoformatter = None

        app = IPython.terminal.ipapp.TerminalIPythonApp.instance(config=config)
        app.initialize(ipython_args)
        shell = app.shell
        for l in lines:
            shell.run_cell(l, silent=True)
        if args.auto_symbols:
            shell.run_cell('from diofant.interactive.session import AutomaticSymbols')
            shell.run_cell('ip = get_ipython()')
            shell.run_cell('ip.ast_transformers.append(AutomaticSymbols(ip.user_ns))')
            shell.run_cell('del ip')
        app.start()
    else:
        ast_transformers = []
        ns = {}

        if not args.no_wrap_division:
            ast_transformers.append(IntegerDivisionWrapper())
        if args.auto_symbols:
            ast_transformers.append(AutomaticSymbols(ns))

        class DiofantConsole(code.InteractiveConsole):
            """An interactive console with readline support."""

            def __init__(self, ast_transformers=[], **kwargs):
                super().__init__(**kwargs)

                readline.set_completer(rlcompleter.Completer(ns).complete)
                readline.parse_and_bind('tab: complete')

                history = os.path.expanduser('~/.python_history')
                readline.read_history_file(history)
                atexit.register(readline.write_history_file, history)
                self.ast_transformers = ast_transformers

            def runsource(self, source, filename='<input>', symbol='single'):
                try:
                    tree = ast.parse(source)
                except SyntaxError:
                    return True

                for t in self.ast_transformers:
                    tree = t.visit(tree)
                ast.fix_missing_locations(tree)

                source = ast.unparse(tree)
                source = source.split('\n')
                source = ';'.join(source)
                return super().runsource(source, filename=filename, symbol=symbol)

        c = DiofantConsole(ast_transformers=ast_transformers, locals=ns)

        for l in lines:
            c.push(l)
        c.interact('', '')


if __name__ == '__main__':  # pragma: no branch
    main()



"""
Some helper functions for the logic module.
"""

import re

from ..core import Symbol
from . import And, Or


def parse_dimacs(s):
    r"""
    Loads a boolean expression from a string in the DIMACS CNF format.

    The file in the DIMACS CNF format is an ASCII file consisting of
    a two major sections: the preamble and the clauses.

    The preamble contains information about the satisfability problem
    instance, which is contained in lines.  Each line begins with a
    single character (followed by whitespace character): either "c" (a
    comment, rest of line is ignored) or "p" (the problem line).
    The problem line has the following format:

        p dnf VARIABLES CLAUSES

    where VARIABLES and CLAUSES fields contain integers, specifying the
    number of variables (n) and the number of clauses (m), respectively.

    The m clauses appear immediately after the problem line.  The variables
    are assumed to be numbered from 1 up to n.  Each clause should be
    represented by a sequence of numbers, separated by one or more
    whitespace characters.  The non-negated version of a variable i is
    represented by i, the negated version is represented by -i.  Each
    clause is terminated by the value 0.

    Examples
    ========

    >>> parse_dimacs('p cnf 2 1\n1 2 0')
    x1 | x2

    References
    ==========

    * https://www.cs.ubc.ca/~hoos/SATLIB/Benchmarks/SAT/satformat.ps

    """
    pComment = re.compile('^c')
    lines = [_ for _ in s.split('\n') if not pComment.match(_)]

    pStats = re.compile(r'^p\s+cnf\s+(?P<nsymbols>\d+)\s+(?P<nclauses>\d+)$')
    m = pStats.match(lines.pop(0))
    if not m:
        raise ValueError('Bad problem line')

    syms = [Symbol(f'x{i}') for i in range(1, int(m.group('nsymbols')) + 1)]
    clauses = []
    nclauses = int(m.group('nclauses'))
    args = []

    while nclauses:
        try:
            line = lines.pop(0)
        except IndexError as exc:
            raise ValueError('Too few clauses') from exc

        line = re.sub(r'^\s*', '', line)
        line = re.sub(r'\s*$', '', line)
        line = re.sub(r'\s+', ' ', line)

        for lit in line.split(' '):
            try:
                ilit = int(lit)
            except ValueError as exc:
                raise ValueError('Bad clause format') from exc
            if ilit == 0:
                clauses.append(Or(*args))
                nclauses -= 1
                args = []
                continue
            num = abs(ilit)
            sign = ilit >= 0
            sym = syms[num - 1]
            args.append(sym if sign else ~sym)

    return And(*clauses)



"""Geometrical Points.

Contains
========
Point
"""

from ..core import Add, Float, Integer, Tuple
from ..core.compatibility import iterable
from ..core.evaluate import global_evaluate
from ..core.sympify import sympify
from ..functions import im, sqrt
from ..matrices import Matrix
from ..simplify import nsimplify, simplify
from ..utilities import ordered
from .entity import GeometryEntity


class Point(GeometryEntity):
    """A point in a n-dimensional Euclidean space.

    Parameters
    ==========

    coords : sequence of n-coordinate values.

    Raises
    ======

    TypeError
        When trying to add or subtract points with different dimensions.
        When `intersection` is called with object other than a Point.

    See Also
    ========

    diofant.geometry.line.Segment : Connects two Points

    Examples
    ========

    >>> Point([1, 2])
    Point(1, 2)
    >>> Point(0, x)
    Point(0, x)

    Floats are automatically converted to Rational unless the
    evaluate flag is False:

    >>> Point(0.5, 0.25)
    Point(1/2, 1/4)
    >>> print(Point(0.5, 0.25, evaluate=False))
    Point(0.5, 0.25)

    """

    def __new__(cls, *args, **kwargs):
        evaluate = kwargs.get('evaluate', global_evaluate[0])

        if iterable(args[0]):
            args = args[0]

        # unpack the arguments into a friendly Tuple
        # if we were already a Point, we're doing an excess
        # iteration, but we'll worry about efficiency later
        coords = Tuple(*args)
        if any(a.is_number and im(a) for a in coords):
            raise ValueError('Imaginary coordinates not permitted.')

        # Turn any Floats into rationals and simplify
        # any expressions before we instantiate
        if evaluate:
            coords = coords.xreplace({f: simplify(nsimplify(f, rational=True))
                                      for f in coords.atoms(Float)})
        if len(coords) != 2:
            raise ValueError('Only two dimensional points currently supported')

        return GeometryEntity.__new__(cls, *coords)

    def __contains__(self, item):
        return item == self

    def is_concyclic(*points):  # pylint: disable=no-method-argument,no-self-argument
        """Is a sequence of points concyclic?

        Test whether or not a sequence of points are concyclic (i.e., they lie
        on a circle).

        Parameters
        ==========

        points : sequence of Points

        Returns
        =======

        is_concyclic : boolean
            True if points are concyclic, False otherwise.

        See Also
        ========

        diofant.geometry.ellipse.Circle

        Notes
        =====

        No points are not considered to be concyclic. One or two points
        are definitely concyclic and three points are conyclic iff they
        are not collinear.

        For more than three points, create a circle from the first three
        points. If the circle cannot be created (i.e., they are collinear)
        then all of the points cannot be concyclic. If the circle is created
        successfully then simply check the remaining points for containment
        in the circle.

        Examples
        ========

        >>> p1, p2 = Point(-1, 0), Point(1, 0)
        >>> p3, p4 = Point(0, 1), Point(-1, 2)
        >>> Point.is_concyclic(p1, p2, p3)
        True
        >>> Point.is_concyclic(p1, p2, p3, p4)
        False

        """
        if not all(isinstance(p, Point) for p in points):
            raise TypeError('Must pass only Point objects')

        if len(points) == 0:
            return False
        if len(points) <= 2:
            return True

        ppoints = list(ordered(Point(p) for p in points))

        from .ellipse import Circle
        try:
            c = Circle(ppoints[0], ppoints[1], ppoints[2])
        except ValueError:
            return False
        for point in ppoints[3:]:
            if point not in c:
                return False
        return True

    def is_collinear(*args):  # pylint: disable=no-method-argument,no-self-argument
        """Is a sequence of points collinear?

        Test whether or not a set of points are collinear. Returns True if
        the set of points are collinear, or False otherwise.

        Parameters
        ==========

        points : sequence of Point

        Returns
        =======

        is_collinear : boolean

        Notes
        =====

        Slope is preserved everywhere on a line, so the slope between
        any two points on the line should be the same. Take the first
        two points, p1 and p2, and create a translated point v1
        with p1 as the origin. Now for every other point we create
        a translated point, vi with p1 also as the origin. Note that
        these translations preserve slope since everything is
        consistently translated to a new origin of p1. Since slope
        is preserved then we have the following equality:

              * v1_slope = vi_slope
              * v1.y/v1.x = vi.y/vi.x (due to translation)
              * v1.y*vi.x = vi.y*v1.x
              * v1.y*vi.x - vi.y*v1.x = 0           (*)

        Hence, if we have a vi such that the equality in (*) is False
        then the points are not collinear. We do this test for every
        point in the list, and if all pass then they are collinear.

        See Also
        ========

        diofant.geometry.line.Line

        Examples
        ========

        >>> p1, p2 = Point(0, 0), Point(1, 1)
        >>> p3, p4, p5 = Point(2, 2), Point(x, x), Point(1, 2)
        >>> Point.is_collinear(p1, p2, p3, p4)
        True
        >>> Point.is_collinear(p1, p2, p3, p5)
        False

        """
        # Coincident points are irrelevant; use only unique points.
        uniq_args = list(set(args))
        if not all(isinstance(p, Point) for p in uniq_args):
            raise TypeError('Must pass only Point objects')

        if len(uniq_args) == 0:
            return False
        if len(uniq_args) <= 2:
            return True

        # translate our points
        points = [p - uniq_args[0] for p in uniq_args[1:]]
        for p in points[1:]:
            if not Point.is_scalar_multiple(points[0], p):
                return False
        return True

    def is_scalar_multiple(self, other):
        """Returns whether `self` and `other` are scalar multiples
        of each other.

        """
        # if the vectors self and other are linearly dependent, then they must
        # be scalar multiples of each other
        m = Matrix([self.args, other.args])
        return m.rank() < 2

    @property
    def length(self):
        """
        Treating a Point as a Line, this returns 0 for the length of a Point.

        Examples
        ========

        >>> p = Point(0, 1)
        >>> p.length
        0

        """
        return Integer(0)

    @property
    def origin(self):
        """A point of all zeros of the same ambient dimension
        as the current point

        """
        return Point([0]*len(self))

    @property
    def is_zero(self):
        """True if every coordinate is zero, otherwise False."""
        return all(x == 0 for x in self.args)

    @property
    def ambient_dimension(self):
        """The dimension of the ambient space the point is in.
        I.e., if the point is in R^n, the ambient dimension
        will be n

        """
        return len(self)

    def distance(self, p):
        """The Euclidean distance from self to point p.

        Parameters
        ==========

        p : Point

        Returns
        =======

        distance : number or symbolic expression.

        See Also
        ========

        diofant.geometry.line.Segment.length

        Examples
        ========

        >>> p1, p2 = Point(1, 1), Point(4, 5)
        >>> p1.distance(p2)
        5

        >>> p3 = Point(x, y)
        >>> p3.distance(Point(0, 0))
        sqrt(x**2 + y**2)

        """
        p = Point(p)
        return sqrt(sum((a - b)**2
                        for a, b in zip(self.args,
                                        p.args if isinstance(p, Point) else p)))

    def midpoint(self, p):
        """The midpoint between self and point p.

        Parameters
        ==========

        p : Point

        Returns
        =======

        midpoint : Point

        See Also
        ========

        diofant.geometry.line.Segment.midpoint

        Examples
        ========

        >>> p1, p2 = Point(1, 1), Point(13, 5)
        >>> p1.midpoint(p2)
        Point(7, 3)

        """
        return Point([simplify((a + b)/2) for a, b in zip(self.args, p.args)])

    def evalf(self, dps=15, **options):
        """Evaluate the coordinates of the point.

        This method will, where possible, create and return a new Point
        where the coordinates are evaluated as floating point numbers to
        the decimal precision dps.

        Returns
        =======

        point : Point

        Examples
        ========

        >>> p1 = Point(Rational(1, 2), Rational(3, 2))
        >>> p1
        Point(1/2, 3/2)
        >>> print(p1.evalf())
        Point(0.5, 1.5)

        """
        coords = [x.evalf(dps, **options) for x in self.args]
        return Point(*coords, evaluate=False)

    def intersection(self, o):
        """The intersection between this point and another point.

        Parameters
        ==========

        other : Point

        Returns
        =======

        intersection : list of Points

        Notes
        =====

        The return value will either be an empty list if there is no
        intersection, otherwise it will contain this point.

        Examples
        ========

        >>> p1, p2, p3 = Point(0, 0), Point(1, 1), Point(0, 0)
        >>> p1.intersection(p2)
        []
        >>> p1.intersection(p3)
        [Point(0, 0)]

        """
        if isinstance(o, Point):
            if self == o:
                return [self]
            return []

        return o.intersection(self)

    def dot(self, p2):
        """Return dot product of self with another Point."""
        p2 = Point(p2)
        return Add(*[a*b for a, b in zip(self, p2)])

    def __len__(self):
        return len(self.args)

    def __iter__(self):
        return self.args.__iter__()

    def __add__(self, other):
        """Add other to self by incrementing self's coordinates by those of other.

        See Also
        ========

        diofant.geometry.entity.GeometryEntity.translate

        """
        if iterable(other) and len(other) == len(self):
            return Point([simplify(a + b) for a, b in zip(self, other)])
        else:
            raise ValueError(
                'Points must have the same number of dimensions')

    def __sub__(self, other):
        """Subtract two points, or subtract a factor from this point's
        coordinates.

        """
        return self + (-other)

    def __mul__(self, factor):
        """Multiply point's coordinates by a factor."""
        factor = sympify(factor)
        return Point([simplify(x*factor) for x in self.args])

    def __truediv__(self, divisor):
        """Divide point's coordinates by a factor."""
        divisor = sympify(divisor)
        return Point([simplify(x/divisor) for x in self.args])

    def __neg__(self):
        """Negate the point."""
        return Point([-x for x in self.args])

    def __abs__(self):
        """Returns the distance between this point and the origin."""
        origin = Point([0]*len(self))
        return Point.distance(origin, self)

    @property
    def x(self):
        """
        Returns the X coordinate of the Point.

        Examples
        ========

        >>> p = Point(0, 1)
        >>> p.x
        0

        """
        return self.args[0]

    @property
    def y(self):
        """
        Returns the Y coordinate of the Point.

        Examples
        ========

        >>> p = Point(0, 1)
        >>> p.y
        1

        """
        return self.args[1]

    @property
    def bounds(self):
        """Return a tuple (xmin, ymin, xmax, ymax) representing the bounding
        rectangle for the geometric figure.

        """
        return self.x, self.y, self.x, self.y

    def rotate(self, angle, pt=None):
        """Rotate ``angle`` radians counterclockwise about Point ``pt``.

        See Also
        ========

        rotate, scale

        Examples
        ========

        >>> t = Point(1, 0)
        >>> t.rotate(pi/2)
        Point(0, 1)
        >>> t.rotate(pi/2, (2, 0))
        Point(2, -1)

        """
        from ..functions import cos, sin

        c = cos(angle)
        s = sin(angle)

        rv = self
        if pt is not None:
            pt = Point(pt)
            rv -= pt
        x, y = rv.args
        rv = Point(c*x - s*y, s*x + c*y)
        if pt is not None:
            rv += pt
        return rv

    def scale(self, x=1, y=1, pt=None):
        """Scale the coordinates of the Point by multiplying by
        ``x`` and ``y`` after subtracting ``pt`` -- default is (0, 0) --
        and then adding ``pt`` back again (i.e. ``pt`` is the point of
        reference for the scaling).

        See Also
        ========

        rotate, translate

        Examples
        ========

        >>> t = Point(1, 1)
        >>> t.scale(2)
        Point(2, 1)
        >>> t.scale(2, 2)
        Point(2, 2)

        """
        if pt:
            pt = Point(pt)
            return self.translate(*(-pt).args).scale(x, y).translate(*pt.args)
        return Point(self.x*x, self.y*y)

    def translate(self, x=0, y=0):
        """Shift the Point by adding x and y to the coordinates of the Point.

        See Also
        ========

        rotate, scale

        Examples
        ========

        >>> t = Point(0, 1)
        >>> t.translate(2)
        Point(2, 1)
        >>> t.translate(2, 2)
        Point(2, 3)
        >>> t + Point(2, 2)
        Point(2, 3)

        """
        return Point(self.x + x, self.y + y)

    def transform(self, matrix):
        """Return the point after applying the transformation described
        by the 3x3 Matrix, ``matrix``.

        See Also
        ========

        diofant.geometry.entity.GeometryEntity.rotate
        diofant.geometry.entity.GeometryEntity.scale
        diofant.geometry.entity.GeometryEntity.translate

        """
        try:
            col, _ = matrix.shape
            valid_matrix = matrix.is_square and col == 3
        except AttributeError:
            # We hit this block if matrix argument is not actually a Matrix.
            valid_matrix = False
        if not valid_matrix:
            raise ValueError('The argument to the transform function must be '
                             + 'a 3x3 matrix')
        x, y = self.args
        return Point(*(Matrix(1, 3, [x, y, 1])*matrix).tolist()[0][:2])



"""Useful utilities for higher level polynomial classes."""

import re

from ..core import Add, Mul, Pow
from ..core.exprtools import decompose_power
from ..utilities import default_sort_key
from .polyerrors import GeneratorsNeeded, PolynomialError
from .polyoptions import build_options


_gens_order = {
    'a': 301, 'b': 302, 'c': 303, 'd': 304,
    'e': 305, 'f': 306, 'g': 307, 'h': 308,
    'i': 309, 'j': 310, 'k': 311, 'l': 312,
    'm': 313, 'n': 314, 'o': 315, 'p': 216,
    'q': 217, 'r': 218, 's': 219, 't': 220,
    'u': 221, 'v': 222, 'w': 223, 'x': 124,
    'y': 125, 'z': 126,
}

_max_order = 1000
_re_gen = re.compile(r'^(.+?)(\d*)$')


def _nsort(roots, separated=False):
    """Sort the numerical roots putting the real roots first, then sorting
    according to real and imaginary parts. If ``separated`` is True, then
    the real and imaginary roots will be returned in two lists, respectively.

    This routine tries to avoid issue sympy/sympy#6137 by separating the roots into real
    and imaginary parts before evaluation. In addition, the sorting will raise
    an error if any computation cannot be done with precision.
    """
    if len(roots) == 1:
        if not separated:
            return list(roots)
        r = list(roots)[0]
        if r.is_extended_real:
            return [[r], []]
        elif r.is_real is False and r.is_complex:
            return [[], [r]]
        else:
            raise NotImplementedError
    all_numbers = all(r.is_number for r in roots)
    if not all_numbers:
        raise NotImplementedError
    # see issue sympy/sympy#6137:
    # get the real part of the evaluated real and imaginary parts of each root
    key = [[i.evalf(2).as_real_imag()[0] for i in r.as_real_imag()] for r in roots]
    # insert a key to indicate if the root has an imaginary part
    key = [(1 if i else 0, r, -abs(i), i.is_positive) for r, i in key]
    key = sorted(zip(key, roots))
    # return the real and imaginary roots separately if desired
    if separated:
        r = []
        i = []
        for (im, _, _, _), v in key:
            if im:
                i.append(v)
            else:
                r.append(v)
        return r, i
    _, roots = zip(*key)
    return list(roots)


def _sort_gens(gens, **args):
    """Sort generators in a reasonably intelligent way."""
    opt = build_options(args)
    gens_order, wrt = _gens_order.copy(), opt.wrt

    for i, gen in enumerate(opt.sort, start=1):
        gens_order[gen] = i

    def order_key(gen):
        gen = str(gen)

        if wrt is not None:
            try:
                return -len(wrt) + wrt.index(gen), gen, 0
            except ValueError:
                pass

        name, index = _re_gen.match(gen).groups()
        index = int(index) if index else 0

        return gens_order.get(name, _max_order), name, index

    return tuple(sorted(gens, key=order_key))


def _unify_gens(f_gens, g_gens):
    """Unify generators in a reasonably intelligent way."""
    f_gens = list(f_gens)
    g_gens = list(g_gens)

    if f_gens == g_gens:
        return tuple(f_gens)

    gens, common, k = [], [], 0

    for gen in f_gens:
        if gen in g_gens:
            common.append(gen)

    for i, gen in enumerate(g_gens):
        if gen in common:
            g_gens[i], k = common[k], k + 1

    for gen in common:
        i = f_gens.index(gen)

        gens.extend(f_gens[:i])
        f_gens = f_gens[i + 1:]

        i = g_gens.index(gen)

        gens.extend(g_gens[:i])
        g_gens = g_gens[i + 1:]

        gens.append(gen)

    gens.extend(f_gens)
    gens.extend(g_gens)

    return tuple(gens)


def _find_gens(exprs, opt):
    """Find generators in a reasonably intelligent way."""
    if opt.domain is not None:
        def _is_coeff(factor):
            return factor in opt.domain
    elif opt.extension is True:
        def _is_coeff(factor):
            return factor.is_number and factor.is_algebraic
    elif opt.greedy is not False:
        def _is_coeff(factor):
            return factor.is_Number and factor.is_finite is not False
    else:
        def _is_coeff(factor):
            return factor.is_number and factor.is_finite is not False

    gens = set()

    for expr in exprs:
        for term in Add.make_args(expr):
            for factor in Mul.make_args(term):
                try:
                    if factor.is_Add and opt.expand:
                        gens |= set(_find_gens([factor], opt))
                    elif not _is_coeff(factor):
                        base, exp = decompose_power(factor)
                        if exp < 0:
                            base = Pow(base, -1)

                        if opt.expand and exp > 1:
                            gens |= set(_find_gens([base], opt))
                        else:
                            gens.add(base)
                except GeneratorsNeeded:
                    pass

    if not gens:
        raise GeneratorsNeeded(f'specify generators to give {exprs} a meaning')

    return _sort_gens(gens, opt=opt)


def _sort_factors(factors, **args):
    """Sort low-level factors in increasing 'complexity' order."""
    def order_if_multiple_key(factor):
        f, n = factor
        return len(f), n, default_sort_key(f)

    def order_no_multiple_key(f):
        return len(f), default_sort_key(f)

    if args.get('multiple', True):
        return sorted(factors, key=order_if_multiple_key)
    else:
        return sorted(factors, key=order_no_multiple_key)


def _parallel_dict_from_expr_if_gens(exprs, opt):
    """Transform expressions into a multinomial form given generators."""
    indices = {g: i for i, g in enumerate(opt.gens)}
    zero_monom = [0]*len(opt.gens)
    polys = []

    for expr in exprs:
        poly = {}

        for term in Add.make_args(expr):
            coeff, monom = [], zero_monom.copy()

            for factor in Mul.make_args(term):
                base, exp = decompose_power(factor)
                if exp < 0:
                    exp, base = -exp, Pow(base, -1)
                try:
                    monom[indices[base]] += exp
                    continue
                except KeyError as exc:
                    if factor.free_symbols & set(opt.gens):
                        raise PolynomialError(f'{factor} contains an element'
                                              ' of the generators set') from exc

                coeff.append(factor)

            monom = tuple(monom)
            poly[monom] = Mul(*coeff) + poly.get(monom, 0)

        polys.append(poly)

    return polys


def parallel_dict_from_expr(exprs, **args):
    """Transform expressions into a multinomial form."""
    reps, opt = _parallel_dict_from_expr(exprs, build_options(args))
    return reps, opt.gens


def _parallel_dict_from_expr(exprs, opt):
    """Transform expressions into a multinomial form."""
    if any(not expr.is_commutative for expr in exprs):
        raise PolynomialError('non-commutative expressions are not supported')

    if opt.expand is not False:
        exprs = [expr.expand() for expr in exprs]

    if not opt.gens:
        opt = opt.clone({'gens': _find_gens(exprs, opt)})

    reps = _parallel_dict_from_expr_if_gens(exprs, opt)

    return reps, opt.clone()



"""
Boolean algebra module for Diofant.
"""

from itertools import combinations, product

from ..core import Atom, cacheit
from ..core.expr import Expr
from ..core.function import Application
from ..core.numbers import Number
from ..core.operations import LatticeOp
from ..core.singleton import S
from ..core.singleton import SingletonWithManagedProperties as Singleton
from ..core.sympify import converter, sympify
from ..utilities import ordered


class Boolean(Expr):
    """A boolean object is an object for which logic operations make sense."""

    def __and__(self, other):
        """Overloading for & operator."""
        return And(self, other)

    __rand__ = __and__

    def __or__(self, other):
        """Overloading for | operator."""
        return Or(self, other)

    __ror__ = __or__

    def __invert__(self):
        """Overloading for ~ operator."""
        return Not(self)

    def __rshift__(self, other):
        """Overloading for >> operator."""
        return Implies(self, other)

    def __lshift__(self, other):
        """Overloading for << operator."""
        return Implies(other, self)

    __rrshift__ = __lshift__
    __rlshift__ = __rshift__

    def __xor__(self, other):
        return Xor(self, other)

    __rxor__ = __xor__

    def equals(self, other, failing_expression=False):
        """
        Returns True if the given formulas have the same truth table.
        For two formulas to be equal they must have the same literals.

        Examples
        ========

        >>> (a >> b).equals(~b >> ~a)
        True
        >>> (~(a & b & c)).equals(~a & ~b & ~c)
        False
        >>> (~(a & ~a)).equals(b | ~b)
        False

        """
        from ..core.relational import Relational
        from .inference import satisfiable

        other = sympify(other)

        if self.has(Relational) or other.has(Relational):
            raise NotImplementedError('handling of relationals')
        return (self.atoms() == other.atoms() and
                not satisfiable(~Equivalent(self, other)))


class BooleanAtom(Atom, Boolean):
    """Base class of BooleanTrue and BooleanFalse."""

    is_Boolean = True

    @property
    def canonical(self):
        return self

    def __int__(self):
        return int(bool(self))


class BooleanTrue(BooleanAtom, metaclass=Singleton):
    """Diofant version of True, a singleton that can be accessed via ``true``.

    This is the Diofant version of True, for use in the logic module. The
    primary advantage of using true instead of True is that shorthand boolean
    operations like ~ and >> will work as expected on this class, whereas with
    True they act bitwise on 1. Functions in the logic module will return this
    class when they evaluate to true.

    Notes
    =====

    There is liable to be some confusion as to when ``True`` should
    be used and when ``true`` should be used in various contexts
    throughout Diofant. An important thing to remember is that
    ``sympify(True)`` returns ``true``. This means that for the most
    part, you can just use ``True`` and it will automatically be converted
    to ``true`` when necessary, similar to how you can generally use 1
    instead of ``Integer(1)``.

    The rule of thumb is:

    "If the boolean in question can be replaced by an arbitrary symbolic
    ``Boolean``, like ``Or(x, y)`` or ``x > 1``, use ``true``.
    Otherwise, use ``True``".

    In other words, use ``true`` only on those contexts where the
    boolean is being used as a symbolic representation of truth.
    For example, if the object ends up in the ``.args`` of any expression,
    then it must necessarily be ``true`` instead of ``True``, as
    elements of ``.args`` must be ``Basic``. On the other hand,
    ``==`` is not a symbolic operation in Diofant, since it always returns
    ``True`` or ``False``, and does so in terms of structural equality
    rather than mathematical, so it should return ``True``. The assumptions
    system should use ``True`` and ``False``. Aside from not satisfying
    the above rule of thumb, the
    assumptions system uses a three-valued logic (``True``, ``False``, ``None``),
    whereas ``true`` and ``false`` represent a two-valued logic. When in
    doubt, use ``True``.

    "``true == True is True``."

    While "``true is True``" is ``False``, "``true == True``"
    is ``True``, so if there is any doubt over whether a function or
    expression will return ``true`` or ``True``, just use ``==``
    instead of ``is`` to do the comparison, and it will work in either
    case.  Finally, for boolean flags, it's better to just use ``if x``
    instead of ``if x is True``. To quote PEP 8:

    Don't compare boolean values to ``True`` or ``False``
    using ``==``.

    * Yes:   ``if greeting:``
    * No:    ``if greeting == True:``
    * Worse: ``if greeting is True:``

    Examples
    ========

    >>> sympify(True)
    true
    >>> ~true
    false
    >>> ~True
    -2
    >>> Or(True, False)
    true

    See Also
    ========

    BooleanFalse

    """

    def __bool__(self):
        return True

    def __hash__(self):
        return hash(True)

    def as_set(self):
        """
        Rewrite logic operators and relationals in terms of real sets.

        Examples
        ========

        >>> true.as_set()
        UniversalSet()

        """
        return S.UniversalSet


class BooleanFalse(BooleanAtom, metaclass=Singleton):
    """Diofant version of False, a singleton that can be accessed via ``false``.

    This is the Diofant version of False, for use in the logic module. The
    primary advantage of using false instead of False is that shorthand boolean
    operations like ~ and >> will work as expected on this class, whereas with
    False they act bitwise on 0. Functions in the logic module will return this
    class when they evaluate to false.

    Notes
    =====

    See note in :py:class:`~diofant.logic.boolalg.BooleanTrue`.

    Examples
    ========

    >>> sympify(False)
    false
    >>> false >> false
    true
    >>> False >> False
    0
    >>> Or(True, False)
    true

    See Also
    ========

    BooleanTrue

    """

    def __bool__(self):
        return False

    def __hash__(self):
        return hash(False)

    def as_set(self):
        """
        Rewrite logic operators and relationals in terms of real sets.

        Examples
        ========

        >>> false.as_set()
        EmptySet()

        """
        from ..sets import EmptySet
        return EmptySet()


true = BooleanTrue()
false: BooleanFalse = BooleanFalse()
# We want S.true and S.false to work, rather than S.BooleanTrue and
# S.BooleanFalse, but making the class and instance names the same causes some
# major issues (like the inability to import the class directly from this
# file).
S.true = true
S.false = false

converter[bool] = lambda x: true if x else false


class BooleanFunction(Application, Boolean):
    """Boolean function is a function that lives in a boolean space.

    This is used as base class for And, Or, Not, etc.

    """

    is_Boolean = True

    def _eval_simplify(self, ratio, measure):
        return simplify_logic(self)


class And(LatticeOp, BooleanFunction):
    """
    Logical AND function.

    It evaluates its arguments in order, giving False immediately
    if any of them are False, and True if they are all True.

    Examples
    ========

    >>> x & y
    x & y

    Notes
    =====

    The ``&`` operator is provided as a convenience, but note that its use
    here is different from its normal use in Python, which is bitwise
    and. Hence, ``And(a, b)`` and ``a & b`` will return different things if
    ``a`` and ``b`` are integers.

    >>> (x & y).subs({x: 1})
    y

    """

    zero = false
    identity = true

    nargs = None

    @classmethod
    def _new_args_filter(cls, args):
        newargs = []
        rel = []
        for x in reversed(list(args)):
            if isinstance(x, Number) or x in (0, 1):
                newargs.append(bool(x))
                continue
            if x.is_Relational:
                c = x.canonical
                if c in rel:
                    continue
                nc = (~c).canonical
                if any(r == nc for r in rel):
                    return [false]
                rel.append(c)
            newargs.append(x)
        return LatticeOp._new_args_filter(newargs, And)

    def as_set(self):
        """
        Rewrite logic operators and relationals in terms of real sets.

        Examples
        ========

        >>> ((x < 2) & (x > -2)).as_set()
        (-2, 2)

        """
        from ..sets import Intersection
        if len(self.free_symbols) == 1:
            return Intersection(*[arg.as_set() for arg in self.args])
        else:
            raise NotImplementedError('Sorry, And.as_set has not yet been'
                                      ' implemented for multivariate'
                                      ' expressions')


class Or(LatticeOp, BooleanFunction):
    """
    Logical OR function

    It evaluates its arguments in order, giving True immediately
    if any of them are True, and False if they are all False.

    Examples
    ========

    >>> x | y
    x | y

    Notes
    =====

    The ``|`` operator is provided as a convenience, but note that its use
    here is different from its normal use in Python, which is bitwise
    or. Hence, ``Or(a, b)`` and ``a | b`` will return different things if
    ``a`` and ``b`` are integers.

    >>> (x | y).subs({x: 0})
    y

    """

    zero = true
    identity = false

    @classmethod
    def _new_args_filter(cls, args):
        newargs = []
        rel = []
        for x in args:
            if isinstance(x, Number) or x in (0, 1):
                newargs.append(bool(x))
                continue
            if x.is_Relational:
                c = x.canonical
                if c in rel:
                    continue
                nc = (~c).canonical
                if any(r == nc for r in rel):
                    return [true]
                rel.append(c)
            newargs.append(x)
        return LatticeOp._new_args_filter(newargs, Or)

    def as_set(self):
        """
        Rewrite logic operators and relationals in terms of real sets.

        Examples
        ========

        >>> ((x > 2) | (x < -2)).as_set()
        [-oo, -2) U (2, oo]

        """
        from ..sets import Union
        if len(self.free_symbols) == 1:
            return Union(*[arg.as_set() for arg in self.args])
        else:
            raise NotImplementedError('Sorry, Or.as_set has not yet been'
                                      ' implemented for multivariate'
                                      ' expressions')


class Not(BooleanFunction):
    """
    Logical Not function (negation).

    Returns True if the statement is False.
    Returns False if the statement is True.

    Examples
    ========

    >>> Not(True)
    false
    >>> Not(False)
    true
    >>> ~And(True, False)
    true
    >>> ~Or(True, False)
    false
    >>> ~(And(True, x) & Or(x, False))
    ~x
    >>> ~x
    ~x
    >>> ~((x | y) & (~x | ~y))
    ~((x | y) & (~x | ~y))

    Notes
    =====

    The ``~`` operator is provided as a convenience, but note that its use
    here is different from its normal use in Python, which is bitwise
    not. In particular, ``~a`` and ``Not(a)`` will be different if ``a`` is
    an integer. Furthermore, since bools in Python subclass from ``int``,
    ``~True`` is the same as ``~1`` which is ``-2``, which has a boolean
    value of True.  To avoid this issue, use the Diofant boolean types
    ``true`` and ``false``.

    >>> ~True
    -2
    >>> ~true
    false

    """

    is_Not = True

    @classmethod
    def eval(cls, arg):
        from ..core import (Equality, GreaterThan, LessThan, StrictGreaterThan,
                            StrictLessThan, Unequality)
        if isinstance(arg, Number) or arg in (True, False):
            return false if arg else true
        if arg.is_Not:
            return arg.args[0]
        # Simplify Relational objects.
        if isinstance(arg, Equality):
            return Unequality(*arg.args)
        if isinstance(arg, Unequality):
            return Equality(*arg.args)
        if isinstance(arg, StrictLessThan):
            return GreaterThan(*arg.args)
        if isinstance(arg, StrictGreaterThan):
            return LessThan(*arg.args)
        if isinstance(arg, LessThan):
            return StrictGreaterThan(*arg.args)
        if isinstance(arg, GreaterThan):
            return StrictLessThan(*arg.args)

    def as_set(self):
        """
        Rewrite logic operators and relationals in terms of real sets.

        Examples
        ========

        >>> Not(x > 0, evaluate=False).as_set()
        (-oo, 0]

        """
        if len(self.free_symbols) == 1:
            return self.args[0].as_set().complement(S.Reals)
        else:
            raise NotImplementedError('Sorry, Not.as_set has not yet been'
                                      ' implemented for mutivariate'
                                      ' expressions')


class Xor(BooleanFunction):
    """
    Logical XOR (exclusive OR) function.

    Returns True if an odd number of the arguments are True and the rest are
    False.

    Returns False if an even number of the arguments are True and the rest are
    False.

    Examples
    ========

    >>> Xor(True, False)
    true
    >>> Xor(True, True)
    false
    >>> Xor(True, False, True, True, False)
    true
    >>> Xor(True, False, True, False)
    false
    >>> x ^ y
    Xor(x, y)

    Notes
    =====

    The ``^`` operator is provided as a convenience, but note that its use
    here is different from its normal use in Python, which is bitwise xor. In
    particular, ``a ^ b`` and ``Xor(a, b)`` will be different if ``a`` and
    ``b`` are integers.

    >>> (x ^ y).subs({y: 0})
    x

    """

    def __new__(cls, *args, **kwargs):
        argset = set()
        obj = super().__new__(cls, *args, **kwargs)
        for arg in super(Xor, obj).args:
            if isinstance(arg, Number) or arg in (True, False):
                if not arg:
                    continue
                arg = true
            if isinstance(arg, Xor):
                for a in arg.args:
                    if a in argset:
                        argset.remove(a)
                    else:
                        argset.add(a)
            elif arg in argset:
                argset.remove(arg)
            else:
                argset.add(arg)
        rel = [(r, r.canonical, (~r).canonical) for r in argset if r.is_Relational]
        odd = False  # is number of complimentary pairs odd? start 0 -> False
        remove = []
        for i, (r, c, nc) in enumerate(rel):
            for j in range(i + 1, len(rel)):
                rj, cj = rel[j][:2]
                if cj == nc:
                    odd = ~odd
                    break
                if cj == c:
                    break
            else:
                continue
            remove.append((r, rj))
        if odd:
            assert true not in argset
            argset.add(true)
        for a, b in remove:
            argset.remove(a)
            argset.remove(b)
        if len(argset) == 0:
            return false
        elif len(argset) == 1:
            return argset.pop()
        elif True in argset:
            argset.remove(True)
            return ~Xor(*argset)
        else:
            obj._args = tuple(ordered(argset))
            obj._argset = frozenset(argset)
            return obj

    @property  # type: ignore[misc]
    @cacheit
    def args(self):
        return tuple(ordered(self._argset))


class Nand(BooleanFunction):
    """
    Logical NAND function.

    It evaluates its arguments in order, giving True immediately if any
    of them are False, and False if they are all True.

    Returns True if any of the arguments are False.
    Returns False if all arguments are True.

    Examples
    ========

    >>> Nand(False, True)
    true
    >>> Nand(True, True)
    false
    >>> Nand(x, y)
    ~(x & y)

    """

    @classmethod
    def eval(cls, *args):
        return ~And(*args)


class Nor(BooleanFunction):
    """
    Logical NOR function.

    It evaluates its arguments in order, giving False immediately if any
    of them are True, and True if they are all False.

    Returns False if any argument is True.
    Returns True if all arguments are False.

    Examples
    ========

    >>> Nor(True, False)
    false
    >>> Nor(True, True)
    false
    >>> Nor(False, True)
    false
    >>> Nor(False, False)
    true
    >>> Nor(x, y)
    ~(x | y)

    """

    @classmethod
    def eval(cls, *args):
        return ~Or(*args)


class Implies(BooleanFunction):
    """
    Logical implication.

    A implies B is equivalent to !A v B

    Accepts two Boolean arguments; A and B.
    Returns False if A is True and B is False.
    Returns True otherwise.

    Examples
    ========

    >>> Implies(True, False)
    false
    >>> Implies(False, False)
    true
    >>> Implies(True, True)
    true
    >>> Implies(False, True)
    true
    >>> x >> y
    Implies(x, y)
    >>> y << x
    Implies(x, y)

    Notes
    =====

    The ``>>`` and ``<<`` operators are provided as a convenience, but note
    that their use here is different from their normal use in Python, which is
    bit shifts. Hence, ``Implies(a, b)`` and ``a >> b`` will return different
    things if ``a`` and ``b`` are integers.  In particular, since Python
    considers ``True`` and ``False`` to be integers, ``True >> True`` will be
    the same as ``1 >> 1``, i.e., 0, which has a truth value of False.  To
    avoid this issue, use the Diofant objects ``true`` and ``false``.

    >>> True >> False
    1
    >>> true >> false
    false

    """

    @classmethod
    def eval(cls, *args):
        try:
            newargs = []
            for x in args:
                if isinstance(x, Number) or x in (0, 1):
                    newargs.append(bool(x))
                else:
                    newargs.append(x)
            A, B = newargs
        except ValueError as exc:
            raise ValueError(f'{len(args)} operand(s) used for an Implies '
                             f'(pairs are required): {args!s}') from exc
        if A == true or A == false or B == true or B == false:
            return Or(Not(A), B)
        elif A == B:
            return true
        elif A.is_Relational and B.is_Relational:
            if A.canonical == B.canonical:
                return true
            elif (~A).canonical == B.canonical:
                return B
        else:
            return Expr.__new__(cls, *args)


class Equivalent(BooleanFunction):
    """
    Equivalence relation.

    Equivalent(A, B) is True iff A and B are both True or both False.

    Returns True if all of the arguments are logically equivalent.
    Returns False otherwise.

    Examples
    ========

    >>> Equivalent(False, False, False)
    true
    >>> Equivalent(True, False, False)
    false
    >>> Equivalent(x, And(x, True))
    true

    """

    def __new__(cls, *args, **options):
        from ..core.relational import Relational
        args = [sympify(arg, strict=True) for arg in args]

        argset = set(args)
        for x in args:
            if isinstance(x, Number) or x in [True, False]:  # Includes 0, 1
                argset.discard(x)
                argset.add(bool(x))
        rel = []
        for r in argset:
            if isinstance(r, Relational):
                rel.append((r, r.canonical, (~r).canonical))
        remove = []
        for i, (r, c, nc) in enumerate(rel):
            for j in range(i + 1, len(rel)):
                rj, cj = rel[j][:2]
                if cj == nc:
                    return false
                elif cj == c:
                    remove.append((r, rj))
                    break
        for a, b in remove:
            argset.remove(a)
            argset.remove(b)
            argset.add(True)
        if len(argset) <= 1:
            return true
        if True in argset:
            argset.discard(True)
            return And(*argset)
        if False in argset:
            argset.discard(False)
            return And(*[~arg for arg in argset])
        _args = frozenset(argset)
        obj = super().__new__(cls, _args)
        obj._argset = _args
        return obj

    @property  # type: ignore[misc]
    @cacheit
    def args(self):
        return tuple(ordered(self._argset))


class ITE(BooleanFunction):
    """
    If then else clause.

    ITE(A, B, C) evaluates and returns the result of B if A is true
    else it returns the result of C.

    Examples
    ========

    >>> ITE(True, False, True)
    false
    >>> ITE(Or(True, False), And(True, True), Xor(True, True))
    true
    >>> ITE(x, y, z)
    ITE(x, y, z)
    >>> ITE(True, x, y)
    x
    >>> ITE(False, x, y)
    y
    >>> ITE(x, y, y)
    y

    """

    @classmethod
    def eval(cls, *args):
        try:
            a, b, c = args
        except ValueError as exc:
            raise ValueError('ITE expects exactly 3 arguments') from exc
        if a == true:
            return b
        elif a == false:
            return c
        elif b == c:
            return b
        elif b == true and c == false:
            return a
        elif b == false and c == true:
            return ~a

    def _eval_derivative(self, x):
        return self.func(self.args[0], *[a.diff(x) for a in self.args[1:]])


def _distribute(expr, a, b):
    """Distributes a over b with respect to expr."""
    if isinstance(expr, b):
        for arg in expr.args:
            if isinstance(arg, a):
                conj = arg
                break
        else:
            return expr
        rest = b(*(a for a in expr.args if a is not conj))
        return a(*(_distribute(b(c, rest), a, b) for c in conj.args))
    elif isinstance(expr, a):
        return a(*(_distribute(c, a, b) for c in expr.args))
    else:
        return expr


def to_nnf(expr, simplify=True):
    """
    Converts expr to Negation Normal Form (NNF).

    If simplify is True, the result contains no redundant clauses.

    Examples
    ========

    >>> to_nnf(~((~a & ~b) | (c & d)))
    (a | b) & (~c | ~d)
    >>> to_nnf(Equivalent(a >> b, b >> a))
    (a | ~b | (a & ~b)) & (b | ~a | (b & ~a))

    See Also
    ========

    is_nnf

    """
    expr = sympify(expr)

    if is_nnf(expr, simplify):
        return expr

    if expr.is_Not:
        expr = expr.args[0]

        if isinstance(expr, And):
            expr = Or(*[~arg for arg in expr.args])
        elif isinstance(expr, Or):
            expr = And(*[~arg for arg in expr.args])
        elif isinstance(expr, Implies):
            a, b = expr.args
            expr = a & ~b
        elif isinstance(expr, Equivalent):
            expr = Or(*expr.args) & Or(*[~arg for arg in expr.args])
        elif isinstance(expr, Xor):
            args = []
            for i in range(1, len(expr.args) + 1, 2):
                for neg in combinations(expr.args, i):
                    args.append(Or(*[~s if s in neg else s for s in expr.args]))
            expr = And(*args)
        elif isinstance(expr, ITE):
            a, b, c = expr.args
            expr = (a | ~c) & (~a | ~b)
        else:
            raise ValueError(f'Illegal operator {expr.func} in expression')

    if isinstance(expr, Implies):
        a, b = expr.args
        expr = ~a | b

    if isinstance(expr, Equivalent):
        args = []
        for a, b in zip(expr.args, expr.args[1:]):
            args.append(~a | b)
        args.append(~expr.args[-1] | expr.args[0])
        expr = And(*args)

    if isinstance(expr, Xor):
        args = []
        for i in range(0, len(expr.args) + 1, 2):
            for neg in combinations(expr.args, i):
                args.append(Or(*[~s if s in neg else s for s in expr.args]))
        expr = And(*args)

    if isinstance(expr, ITE):
        a, b, c = expr.args
        expr = (~a | b) & (a | c)

    args = []
    for arg in expr.args:
        if not is_literal(arg):
            arg = to_nnf(arg, simplify)
        if simplify:
            arg = arg.args if isinstance(arg, expr.func) else (arg,)
            for a in arg:
                if ~a in args:
                    return expr.func.zero
                args.append(a)
        else:
            args.append(arg)

    return expr.func(*args)


def to_cnf(expr, simplify=False):
    """
    Convert expr to Conjunctive Normal Form (CNF).

    If simplify is True, the expr is evaluated to its simplest CNF form.

    Examples
    ========

    >>> to_cnf(~(a | b) | c)
    (c | ~a) & (c | ~b)
    >>> to_cnf((a | b) & (a | ~a), True)
    a | b

    See Also
    ========

    is_cnf

    """
    expr = sympify(expr)

    if not isinstance(expr, BooleanFunction):
        return expr

    if simplify:
        return simplify_logic(expr, 'cnf')

    if is_cnf(expr):
        return expr

    return _distribute(to_nnf(expr), And, Or)


def to_dnf(expr, simplify=False):
    """
    Convert expr to Disjunctive Normal Form (DNF).

    If simplify is True, the expr is evaluated to its simplest DNF form.

    Examples
    ========

    >>> to_dnf(b & (a | c))
    (a & b) | (b & c)
    >>> to_dnf((a & b) | (a & ~b) | (b & c) | (~b & c), True)
    a | c

    See Also
    ========

    is_dnf

    """
    expr = sympify(expr)

    if not isinstance(expr, BooleanFunction):
        return expr

    if simplify:
        return simplify_logic(expr, 'dnf')

    if is_dnf(expr):
        return expr

    return _distribute(to_nnf(expr), Or, And)


def is_nnf(expr, simplified=True):
    """
    Checks if expr is in Negation Normal Form (NNF).

    A logical expression is in NNF if the negation operator is only
    applied to literals and the only other allowed boolean functions
    are conjunction and disjunction.

    If simplified is True, checks if result contains no redundant clauses.

    Examples
    ========

    >>> is_nnf(a & b | ~c)
    True
    >>> is_nnf((a | ~a) & (b | c))
    False
    >>> is_nnf((a | ~a) & (b | c), False)
    True
    >>> is_nnf(~(a & b) | c)
    False
    >>> is_nnf((a >> b) & (b >> a))
    False

    See Also
    ========

    to_nnf
    is_cnf
    is_dnf

    """
    expr = sympify(expr)
    if is_literal(expr):
        return True

    stack = [expr]

    while stack:
        expr = stack.pop()
        if expr.func in (And, Or):
            if simplified:
                args = expr.args
                for arg in args:
                    if ~arg in args:
                        return False
            stack.extend(expr.args)

        elif not is_literal(expr):
            return False

    return True


def is_cnf(expr):
    """
    Checks if expr is in Conjunctive Normal Form (CNF).

    A logical expression is in CNF if it is a conjunction of one or more
    clauses, where a clause is a disjunction of literals.

    Examples
    ========

    >>> is_cnf(a | b | c)
    True
    >>> is_cnf(a & b & c)
    True
    >>> is_cnf((a & b) | c)
    False

    See Also
    ========

    to_cnf
    is_dnf
    is_nnf

    """
    return _is_form(expr, And, Or)


def is_dnf(expr):
    """
    Checks if expr is in Disjunctive Normal Form (DNF).

    A logical expression is in DNF if it is a disjunction of one or more
    clauses, where a clause is a conjunction of literals.

    Examples
    ========

    >>> is_dnf(a | b | c)
    True
    >>> is_dnf(a & b & c)
    True
    >>> is_dnf((a & b) | c)
    True
    >>> is_dnf(a & (b | c))
    False

    See Also
    ========

    to_dnf
    is_cnf
    is_nnf

    """
    return _is_form(expr, Or, And)


def _is_form(expr, function1, function2):
    expr = sympify(expr)

    if expr.is_Atom:
        return True

    if isinstance(expr, function2):
        for lit in expr.args:
            if isinstance(lit, Not):
                if not lit.args[0].is_Atom:
                    return False
            else:
                if not lit.is_Atom:
                    return False
        return True

    if isinstance(expr, Not):
        if not expr.args[0].is_Atom:
            return False

    if not isinstance(expr, function1):
        return False

    for cls in expr.args:
        if cls.is_Atom:
            continue
        if isinstance(cls, Not):
            if not cls.args[0].is_Atom:
                return False
        elif not isinstance(cls, function2):
            return False
        for lit in cls.args:
            if isinstance(lit, Not):
                if not lit.args[0].is_Atom:
                    return False
            else:
                if not lit.is_Atom:
                    return False

    return True


def is_literal(expr):
    """
    Returns True if expr is a literal, else False.

    Examples
    ========

    >>> is_literal(a)
    True
    >>> is_literal(~a)
    True
    >>> is_literal(a + b)
    True
    >>> is_literal(a | b)
    False

    """
    if isinstance(expr, Not):
        expr = expr.args[0]
    return not isinstance(expr, BooleanFunction)


def to_int_repr(clauses, symbols):
    """
    Takes clauses in CNF format and puts them into an integer representation.

    Examples
    ========

    >>> to_int_repr([x | y, y], [x, y])
    [{1, 2}, {2}]

    """
    symbols = dict(zip(symbols, range(1, len(symbols) + 1)))

    def append_symbol(arg, symbols):
        if isinstance(arg, Not):
            return -symbols[arg.args[0]]
        else:
            return symbols[arg]

    return [{append_symbol(arg, symbols) for arg in Or.make_args(c)}
            for c in clauses]


def _check_pair(minterm1, minterm2):
    """
    Checks if a pair of minterms differs by only one bit. If yes, returns
    index, else returns -1.

    """
    index = -1
    for x, (i, j) in enumerate(zip(minterm1, minterm2)):
        if i != j:
            if index == -1:
                index = x
            else:
                return -1
    return index


def _convert_to_varsSOP(minterm, variables):
    """
    Converts a term in the expansion of a function from binary to it's
    variable form (for SOP).

    """
    temp = []
    for i, m in enumerate(minterm):
        if m == 0:
            temp.append(~variables[i])
        elif m == 1:
            temp.append(variables[i])
    return And(*temp)


def _convert_to_varsPOS(maxterm, variables):
    """
    Converts a term in the expansion of a function from binary to it's
    variable form (for POS).

    """
    temp = []
    for i, m in enumerate(maxterm):
        if m == 1:
            temp.append(~variables[i])
        elif m == 0:
            temp.append(variables[i])
    return Or(*temp)


def _simplified_pairs(terms):
    """
    Reduces a set of minterms, if possible, to a simplified set of minterms
    with one less variable in the terms using QM method.

    """
    simplified_terms = []
    todo = list(range(len(terms)))
    for i, ti in enumerate(terms[:-1]):
        for j_i, tj in enumerate(terms[(i + 1):]):
            index = _check_pair(ti, tj)
            if index != -1:
                todo[i] = todo[j_i + i + 1] = None
                newterm = ti[:]
                newterm[index] = 3
                if newterm not in simplified_terms:
                    simplified_terms.append(newterm)
    simplified_terms.extend(
        [terms[i] for i in [_ for _ in todo if _ is not None]])
    return simplified_terms


def _compare_term(minterm, term):
    """
    Return True if a binary term is satisfied by the given term. Used
    for recognizing prime implicants.

    """
    for i, x in enumerate(term):
        if x not in (3, minterm[i]):
            return False
    return True


def _rem_redundancy(l1, terms):
    """
    After the truth table has been sufficiently simplified, use the prime
    implicant table method to recognize and eliminate redundant pairs,
    and return the essential arguments.

    """
    essential = []
    for x in terms:
        temporary = []
        for y in l1:
            if _compare_term(x, y):
                temporary.append(y)
        if len(temporary) == 1:
            if temporary[0] not in essential:
                essential.append(temporary[0])
    for x in terms:
        for y in essential:
            if _compare_term(x, y):
                break
        else:
            for z in l1:  # pragma: no branch
                if _compare_term(x, z):
                    assert z not in essential
                    essential.append(z)
                    break

    return essential


def _SOPform(variables, minterms, dontcares=[]):
    """
    The _SOPform function uses simplified_pairs and a redundant group-
    eliminating algorithm to convert the list of all input combos that
    generate '1' (the minterms) into the smallest Sum of Products form.

    The variables must be given as the first argument.

    Return a logical Or function (i.e., the "sum of products" or "SOP"
    form) that gives the desired outcome. If there are inputs that can
    be ignored, pass them as a list, too.

    The result will be one of the (perhaps many) functions that satisfy
    the conditions.

    Examples
    ========

    >>> minterms = [[0, 0, 0, 1], [0, 0, 1, 1],
    ...             [0, 1, 1, 1], [1, 0, 1, 1], [1, 1, 1, 1]]
    >>> dontcares = [[0, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 1]]
    >>> _SOPform([t, x, y, z], minterms, dontcares)
    (y & z) | (z & ~t)

    References
    ==========

    * https://en.wikipedia.org/wiki/Quine-McCluskey_algorithm

    """
    variables = [sympify(v) for v in variables]
    if minterms == []:
        return false

    minterms = [list(i) for i in minterms]
    dontcares = [list(i) for i in dontcares]
    for d in dontcares:
        if d in minterms:
            raise ValueError(f'{d} in minterms is also in dontcares')

    old = None
    new = minterms + dontcares
    while new != old:
        old = new
        new = _simplified_pairs(old)
    essential = _rem_redundancy(new, minterms)
    return Or(*[_convert_to_varsSOP(x, variables) for x in essential])


def _POSform(variables, minterms, dontcares=[]):
    """
    The _POSform function uses simplified_pairs and a redundant-group
    eliminating algorithm to convert the list of all input combinations
    that generate '1' (the minterms) into the smallest Product of Sums form.

    The variables must be given as the first argument.

    Return a logical And function (i.e., the "product of sums" or "POS"
    form) that gives the desired outcome. If there are inputs that can
    be ignored, pass them as a list, too.

    The result will be one of the (perhaps many) functions that satisfy
    the conditions.

    Examples
    ========

    >>> minterms = [[0, 0, 0, 1], [0, 0, 1, 1], [0, 1, 1, 1],
    ...             [1, 0, 1, 1], [1, 1, 1, 1]]
    >>> dontcares = [[0, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 1]]
    >>> _POSform([t, x, y, z], minterms, dontcares)
    z & (y | ~t)

    References
    ==========

    * https://en.wikipedia.org/wiki/Quine-McCluskey_algorithm

    """
    variables = [sympify(v) for v in variables]
    if minterms == []:
        return false

    minterms = [list(i) for i in minterms]
    dontcares = [list(i) for i in dontcares]
    for d in dontcares:
        if d in minterms:
            raise ValueError(f'{d} in minterms is also in dontcares')

    maxterms = []
    for t in product([0, 1], repeat=len(variables)):
        t = list(t)
        if (t not in minterms) and (t not in dontcares):
            maxterms.append(t)
    old = None
    new = maxterms + dontcares
    while new != old:
        old = new
        new = _simplified_pairs(old)
    essential = _rem_redundancy(new, maxterms)
    return And(*[_convert_to_varsPOS(x, variables) for x in essential])


def _find_predicates(expr):
    """Helper to find logical predicates in BooleanFunctions.

    A logical predicate is defined here as anything within a BooleanFunction
    that is not a BooleanFunction itself.

    """
    if not isinstance(expr, BooleanFunction):
        return {expr}
    return set().union(*(_find_predicates(i) for i in expr.args))


def simplify_logic(expr, form='cnf', deep=True):
    """
    This function simplifies a boolean function to its simplified version
    in SOP or POS form. The return type is an Or or And object in Diofant.

    Parameters
    ==========

    expr : string or boolean expression
    form : string ('cnf' or 'dnf'), default to 'cnf'.
        Selects the normal form in which the result is returned.
    deep : boolean (default True)
        indicates whether to recursively simplify any
        non-boolean functions contained within the input.

    Examples
    ========

    >>> b = (~x & ~y & ~z) | (~x & ~y & z)
    >>> simplify_logic(b)
    ~x & ~y

    >>> sympify(b)
    (z & ~x & ~y) | (~x & ~y & ~z)
    >>> simplify_logic(_)
    ~x & ~y

    """
    if form in ('cnf', 'dnf'):
        expr = sympify(expr)
        variables = _find_predicates(expr)
        truthtable = []
        for t in product([0, 1], repeat=len(variables)):
            t = list(t)
            if expr.xreplace(dict(zip(variables, t))):
                truthtable.append(t)
        if deep:
            variables = [v.simplify() for v in variables]
        if form == 'dnf':
            return _SOPform(variables, truthtable)
        else:
            return _POSform(variables, truthtable)
    else:
        raise ValueError('form can be cnf or dnf only')



"""Tests for modular GCD algorithms."""

from diofant import QQ, ZZ, ring, sqrt
from diofant.polys.modulargcd import (_chinese_remainder_reconstruction,
                                      _func_field_modgcd_m, _to_ANP_poly,
                                      _to_ZZ_poly)


__all__ = ()


def test_chinese_remainder():
    R, x, y = ring('x y', ZZ)
    p, q = 3, 5

    hp = x**3*y - x**2 - 1
    hq = -x**3*y - 2*x*y**2 + 2

    hpq = _chinese_remainder_reconstruction(hp, hq, p, q)

    assert hpq.trunc_ground(p) == hp
    assert hpq.trunc_ground(q) == hq

    _, z = ring('z', R)
    p, q = 3, 7

    hp = (x*y + 1)*z**2 + x
    hq = (x**2 - 3*y)*z + 2

    hpq = _chinese_remainder_reconstruction(hp, hq, p, q)

    assert hpq.trunc_ground(p) == hp
    assert hpq.trunc_ground(q) == hq


def test_to_ZZ_ANP_poly():
    A = QQ.algebraic_field(sqrt(2))
    R, x = ring('x', A)
    f = x*(sqrt(2) + 1)

    T, x_, z_ = ring('x_ z_', ZZ)
    f_ = x_*z_ + x_

    assert _to_ZZ_poly(f, T) == f_
    assert _to_ANP_poly(f_, R) == f

    R, x, t, s = ring('x t s', A)
    f = x*t**2 + x*s + sqrt(2)

    D, t_, s_ = ring('t_ s_', ZZ)
    T, x_, z_ = ring('x_ z_', D)
    f_ = (t_**2 + s_)*x_ + z_

    assert _to_ZZ_poly(f, T) == f_
    assert _to_ANP_poly(f_, R) == f


def test_modgcd_func_field():
    D, t = ring('t', ZZ)
    _, x, z = ring('x z', D)

    minpoly = (z**2*t**2 + z**2*t - 1).drop(0)
    f, g = x + 1, x - 1

    assert _func_field_modgcd_m(f, g, minpoly) == 1

    # First example from Monagan2004algebraic.
    m = z**2 - t
    f = 3*t*x**2 - (2*t**2 - 3*t)*x*z + 15*x + 15*z - 2*t**3
    g = 3*t*x**2*z + 15*x*z + (-2*t**3 + 3*t)*x - 2*t**2*z + 15

    assert _func_field_modgcd_m(f, g, m.drop(0)) == 3*t*x - 2*t**2*z + 15

    g = 3*t*x - 2*t**2*z + 15
    a = x + z
    b = x*z + 1

    assert _func_field_modgcd_m(a*g, b*g, m.drop(0)) == g % m
    assert _func_field_modgcd_m(a*g**2, b*g**2, m.drop(0)) == g**2 % m

    # issue diofant/diofant#850
    assert _func_field_modgcd_m(a*g**3, b*g**3, m.drop(0)) == g**3 % m



import pytest

from diofant import (Derivative, E, I, O, PoleError, Rational, Symbol, acosh,
                     acoth, asin, asinh, atanh, besselk, cbrt, ceiling, cos,
                     cosh, cot, coth, exp, floor, limit, ln, log, oo, pi, sign,
                     sin, sinh, sqrt, tan, tanh)
from diofant.abc import a, b, l, w, x, y, z


__all__ = ()


def test_simple_1():
    assert x.series(x, n=5) == x
    assert y.series(x, n=5) == y
    assert (1/(x*y)).series(y, n=5) == 1/(x*y)
    assert Rational(3, 4).series(x, n=5) == Rational(3, 4)
    assert x.series(x) == x

    # issue sympy/sympy#5183
    assert ((1 + x)**2).series(x) == 1 + 2*x + x**2
    assert (1 + 1/x).series() == 1 + 1/x
    assert (Derivative(exp(x).series(), x).doit() ==
            1 + x + x**2/2 + x**3/6 + x**4/24 + Derivative(O(x**6), x))


def test_mul_0():
    assert (x*ln(x)).series(x, n=5) == x*ln(x)


def test_mul_1():
    assert (x*ln(2 + x)).series(x, n=5) == x*log(2) + x**2/2 - x**3/8 + \
        x**4/24 + O(x**5)
    assert (x*ln(1 + x)).series(x, n=5) == x**2 - x**3/2 + x**4/3 + O(x**5)


def test_pow_0():
    assert (x**2).series(x, n=5) == x**2
    assert (1/x).series(x, n=5) == 1/x
    assert (1/x**2).series(x, n=5) == 1/x**2
    assert (x**Rational(2, 3)).series(x, n=5) == (x**Rational(2, 3))
    assert (sqrt(x)**3).series(x, n=5) == (sqrt(x)**3)


def test_pow_1():
    assert ((1 + x)**2).series(x, n=5) == 1 + 2*x + x**2


def test_geometric_1():
    assert (1/(1 - x)).series(x, n=5) == 1 + x + x**2 + x**3 + x**4 + O(x**5)
    assert (x/(1 - x)).series(x, n=7) == x + x**2 + x**3 + x**4 + x**5 + \
        x**6 + O(x**7)
    assert (x**3/(1 - x)).series(x, n=8) == x**3 + x**4 + x**5 + x**6 + \
        x**7 + O(x**8)


def test_sqrt_1():
    assert sqrt(1 + x).series(x, n=5) == 1 + x/2 - x**2/8 + x**3/16 - 5*x**4/128 + O(x**5)


def test_exp_1():
    assert exp(x).series(x, n=5) == 1 + x + x**2/2 + x**3/6 + x**4/24 + O(x**5)
    assert exp(x).series(x, n=12) == 1 + x + x**2/2 + x**3/6 + x**4/24 + x**5/120 +  \
        x**6/720 + x**7/5040 + x**8/40320 + x**9/362880 + x**10/3628800 +  \
        x**11/39916800 + O(x**12)
    assert exp(1/x).series(x, n=5) == exp(1/x)
    assert exp(1/(1 + x)).series(x, n=4) ==  \
        (E*(1 - x - 13*x**3/6 + 3*x**2/2)).expand() + O(x**4)
    assert exp(2 + x).series(x, n=5) ==  \
        (exp(2)*(1 + x + x**2/2 + x**3/6 + x**4/24)).expand() + O(x**5)


def test_exp_sqrt_1():
    assert exp(1 + sqrt(x)).series(x, n=2) == \
        E + E*x/2 + E*sqrt(x) + E*x**Rational(3, 2)/6 + O(x**2)


def test_power_x_x1():
    assert (exp(x*ln(x))).series(x, n=3) == \
        1 + x*log(x) + x**2*log(x)**2/2 + x**3*log(x)**3/6 + O(x**3)


def test_power_x_x2():
    assert (x**x).series(x, n=3) == \
        1 + x*log(x) + x**2*log(x)**2/2 + x**3*log(x)**3/6 + O(x**3)


def test_log_singular1():
    assert log(1 + 1/x).series(x, n=5) == x - log(x) - x**2/2 + x**3/3 - \
        x**4/4 + O(x**5)


def test_log_power1():
    e = 1 / (1/x + x ** (log(3)/log(2)))
    assert e.series(x, n=3) == x - x**(2 + log(3)/log(2)) + O(x**3)


def test_log_series():
    e = 1/(1 - log(x))
    assert e.series(x, n=5, logx=l) == 1/(1 - l)


def test_log2():
    e = log(-1/x)
    assert e.series(x, n=5) == -log(x) + log(-1)


def test_log3():
    e = 1/log(-1/x)
    assert e.series(x, n=4, logx=l) == 1/(-l + log(-1))


def test_log4():
    assert (log(1 + x).series(x, x0=I*oo) ==
            1/(5*x**5) - 1/(4*x**4) + 1/(3*x**3) - 1/(2*x**2) + 1/x +
            I*pi/2 + log(-I*x) + O(x**(-6), (x, oo*I)))


def test_x0():
    # issue sympy/sympy#5654
    assert ((1/(x**2 + a**2)**2).series(x, x0=I*a, n=0) ==
            -I/(4*a**3*(-I*a + x)) - 1/(4*a**2*(-I*a + x)**2) + O(1, (x, I*a)))
    assert ((1/(x**2 + a**2)**2).series(x, x0=I*a, n=1) ==
            3/(16*a**4) - I/(4*a**3*(-I*a + x)) - 1/(4*a**2*(-I*a + x)**2) +
            O(-I*a + x, (x, I*a)))


def test_series1():
    e = sin(x)
    assert e.series(x, n=0) != 0
    assert e.series(x, n=1) == O(x)
    assert e.series(x, n=2) == x + O(x**2)
    assert e.series(x, n=3) == x + O(x**3)
    assert e.series(x, n=5) == x - x**3/6 + O(x**5)

    e = (exp(x) - 1)/x
    assert e.series(x, n=3) == 1 + x/2 + x**2/6 + O(x**3)

    assert x.series(x, n=2) == x


def test_series1_failing():
    assert x.series(x, n=0) == O(1, x)
    assert x.series(x, n=1) == O(x)


def test_seriesbug1():
    assert (1/x).series(x, n=3) == 1/x
    assert (x + 1/x).series(x, n=3) == x + 1/x


def test_series2x():
    assert ((x + 1)**(-2)).series(x, n=4) == 1 - 2*x + 3*x**2 - 4*x**3 + O(x**4)
    assert ((x + 1)**(-1)).series(x, n=4) == 1 - x + x**2 - x**3 + O(x**4)
    assert ((x + 1)**0).series(x, n=3) == 1
    assert ((x + 1)**1).series(x, n=3) == 1 + x
    assert ((x + 1)**2).series(x, n=3) == 1 + 2*x + x**2
    assert ((x + 1)**3).series(x, n=3) == 1 + 3*x + 3*x**2 + O(x**3)

    assert (1/(1 + x)).series(x, n=4) == 1 - x + x**2 - x**3 + O(x**4)
    assert (x + 3/(1 + 2*x)).series(x, n=4) == 3 - 5*x + 12*x**2 - 24*x**3 + O(x**4)

    assert ((1/x + 1)**3).series(x, n=4) == 1 + x**(-3) + 3*x**(-2) + 3/x
    assert (1/(1 + 1/x)).series(x, n=4) == x - x**2 + x**3 - O(x**4)
    assert (1/(1 + 1/x**2)).series(x) == x**2 - x**4 + O(x**6)


def test_bug2():  # 1/log(0) * log(0) problem
    e = (w**(-1) + w**(
        -log(3)*log(2)**(-1)))**(-1)*(3*w**(-log(3)*log(2)**(-1)) + 2*w**(-1))
    e = e.expand()
    assert e.series(w, n=4).subs({w: 0}) == 3


def test_exp():
    e = (1 + x)**(1/x)
    assert e.series(x, n=2) == exp(1) - x*exp(1)/2 + O(x**2)


def test_exp2():
    e = w**(1 - log(x)/(log(2) + log(x)))
    assert e.series(w, n=1) == e


def test_bug3():
    e = (2/x + 3/x**2)/(1/x + 1/x**2)
    assert e.series(x, n=3) == 3 - x + x**2 + O(x**3)


def test_generalexponent():
    p = 2
    e = (2/x + 3/x**p)/(1/x + 1/x**p)
    assert e.series(x, n=3) == 3 - x + x**2 + O(x**3)
    p = Rational(1, 2)
    e = (2/x + 3/x**p)/(1/x + 1/x**p)
    assert e.series(x, n=2) == 2 - x + sqrt(x) + x**Rational(3, 2) + O(x**2)

    e = 1 + sqrt(x)
    assert e.series(x, n=4) == 1 + sqrt(x)

# more complicated example


def test_genexp_x():
    e = 1/(1 + sqrt(x))
    assert e.series(x, n=2) == 1 + x - sqrt(x) - sqrt(x)**3 + O(x**2)

# more complicated example


def test_genexp_x2():
    p = Rational(3, 2)
    e = (2/x + 3/x**p)/(1/x + 1/x**p)
    assert e.series(x, n=2) == 3 + x - sqrt(x) - x**p + O(x**2)


def test_seriesbug2():
    # simple case (1):
    e = ((2*w)/w)**(1 + w)
    assert e.series(w, n=1) == 2 + O(w)
    assert e.series(w, n=1).subs({w: 0}) == 2


def test_seriesbug2b():
    # test sin
    e = sin(2*w)/w
    assert e.series(w, n=2) == 2 + O(w**2)


def test_seriesbug2d():
    w = Symbol('w', extended_real=True)
    e = log(sin(2*w)/w)
    assert e.series(w, n=5) == log(2) - 2*w**2/3 - 4*w**4/45 + O(w**5)


def test_seriesbug2c():
    w = Symbol('w', extended_real=True)
    # more complicated case, but sin(x)~x, so the result is the same as in (1)
    e = (sin(2*w)/w)**(1 + w)
    assert e.series(w, 0, 1) == 2 + O(w)
    assert e.series(w, 0, 3) == 2 + 2*w*log(2) + \
        w**2*(-Rational(4, 3) + log(2)**2) + O(w**3)
    assert e.series(w, 0, 2).subs({w: 0}) == 2


def test_expbug4():
    x = Symbol('x', extended_real=True)
    assert (log(
        sin(2*x)/x)*(1 + x)).series(x, 0, 2) == log(2) + x*log(2) + O(x**2)
    assert exp(
        log(sin(2*x)/x)*(1 + x)).series(x, 0, 2) == 2 + 2*x*log(2) + O(x**2)

    assert exp(log(2) + O(x)).series(x, n=1) == 2 + O(x)
    assert ((2 + O(x))**(1 + x)).series(x, n=1) == 2 + O(x)


def test_logbug4():
    assert log(2 + O(x)).series(x, n=1) == log(2) + O(x)


def test_expbug5():
    assert exp(log(1 + x)/x).series(x, n=2) == exp(1) + -exp(1)*x/2 + O(x**2)

    assert exp(O(x)).series(x, n=1) == 1 + O(x)


def test_sinsinbug():
    assert sin(sin(x)).series(x, n=9) == x - x**3/3 + x**5/10 - 8*x**7/315 + O(x**9)


def test_sympyissue_3258():
    a = x/(exp(x) - 1)
    assert a.series(x, n=5) == 1 - x/2 - x**4/720 + x**2/12 + O(x**5)


def test_sympyissue_3204():
    x = Symbol('x', nonnegative=True)
    f = cbrt(sin(x**3))
    assert f.series(x, n=19) == x - x**7/18 - x**13/3240 + O(x**19)


def test_sympyissue_3224():
    f = sqrt(1 - sqrt(y))
    assert f.series(y, n=2) == 1 - sqrt(y)/2 - y/8 - sqrt(y)**3/16 + O(y**2)


def test_sympyissue_3463():
    r = log(5)/log(3)
    p = w**(-1 + r)
    e = 1/x*(-log(w**(1 + r)) + log(w + w**r))
    e_ser = -r*log(w)/x + p/x - p**2/(2*x) + O(p**3)
    assert (e.series(w, n=3) - e_ser).removeO().simplify() == 0


def test_sin():
    assert sin(8*x).series(x, n=4) == 8*x - 256*x**3/3 + O(x**4)
    assert sin(x + y).series(x, n=1) == sin(y) + O(x)
    assert sin(x + y).series(x, n=2) == sin(y) + cos(y)*x + O(x**2)
    assert sin(x + y).series(x, n=5) == sin(y) + cos(y)*x - sin(y)*x**2/2 - \
        cos(y)*x**3/6 + sin(y)*x**4/24 + O(x**5)


def test_sympyissue_3515():
    e = sin(8*x)/x
    assert e.series(x) == 8 - 256*x**2/3 + 4096*x**4/15 + O(x**6)


def test_sympyissue_3505():
    e = sin(x)**(-4)*(sqrt(cos(x))*sin(x)**2 - cbrt(cos(x))*sin(x)**2)
    assert e.series(x) == -Rational(1, 12) - 7*x**2/288 - \
        43*x**4/10368 + O(x**6)


def test_sympyissue_3501():
    e = x**(-2)*(x*sin(a + x) - x*sin(a))
    assert e.series(x, n=4) == cos(a) - sin(a)*x/2 - cos(a)*x**2/6 + \
        sin(a)*x**3/24 + O(x**4)
    e = x**(-2)*(x*cos(a + x) - x*cos(a))
    assert e.series(x, n=5) == -sin(a) - cos(a)*x/2 + sin(a)*x**2/6 + \
        cos(a)*x**3/24 - x**4*sin(a)/120 + O(x**5)


def test_sympyissue_3502():
    e = sin(5*x)/sin(2*x)
    assert e.series(x, n=2) == Rational(5, 2) + O(x**2)
    assert e.series(x) == \
        Rational(5, 2) - 35*x**2/4 + 329*x**4/48 + O(x**6)


def test_sympyissue_3503():
    e = sin(2 + x)/(2 + x)
    assert e.series(x, n=2) == sin(2)/2 + x*(-sin(2)/4 + cos(2)/2) + O(x**2)


def test_sympyissue_3506():
    e = (x + sin(3*x))**(-2)*(x*(x + sin(3*x)) - (x + sin(3*x))*sin(2*x))
    assert e.series(x) == -Rational(1, 4) + 5*x**2/96 + 91*x**4/768 + O(x**6)


def test_sympyissue_3508():
    x = Symbol('x', extended_real=True)
    assert log(sin(x)).series(x, n=5) == log(x) - x**2/6 - x**4/180 + O(x**5)
    e = -log(x) + x*(-log(x) + log(sin(2*x))) + log(sin(2*x))
    assert e.series(x, n=5) == \
        log(2) + log(2)*x - 2*x**2/3 - 2*x**3/3 - 4*x**4/45 + O(x**5)


def test_sympyissue_3507():
    e = x**(-4)*(x**2 - x**2*sqrt(cos(x)))
    assert e.series(x) == \
        Rational(1, 4) + x**2/96 + 19*x**4/5760 + O(x**6)


def test_sympyissue_3639():
    assert sin(cos(x)).series(x, n=5) == \
        sin(1) - x**2*cos(1)/2 + x**4*(-sin(1)/8 + cos(1)/24) + O(x**5)


def test_hyperbolic():
    assert sinh(x).series(x, n=7) == x + x**3/6 + x**5/120 + O(x**7)
    assert cosh(x).series(x) == 1 + x**2/2 + x**4/24 + O(x**6)
    assert tanh(x).series(x, n=7) == x - x**3/3 + 2*x**5/15 + O(x**7)
    assert coth(x).series(x, n=7) == \
        1/x - x**3/45 + x/3 + 2*x**5/945 + O(x**7)
    assert asinh(x).series(x, n=7) == x - x**3/6 + 3*x**5/40 + O(x**7)
    assert acosh(x).series(x, n=7) == \
        pi*I/2 - I*x - 3*I*x**5/40 - I*x**3/6 + O(x**7)
    assert atanh(x).series(x, n=7) == x + x**3/3 + x**5/5 + O(x**7)
    assert acoth(x).series(x, n=7) == -I*pi/2 + x + x**3/3 + x**5/5 + O(x**7)


def test_series2():
    w = Symbol('w', extended_real=True)
    x = Symbol('x', extended_real=True)
    e = w**(-2)*(w*exp(1/x - w) - w*exp(1/x))
    assert e.series(w, n=2) == -exp(1/x) + w*exp(1/x)/2 + O(w**2)


def test_series3():
    w = Symbol('w', extended_real=True)
    e = w**(-6)*(w**3*tan(w) - w**3*sin(w))
    assert e.series(w, n=2) == Rational(1, 2) + O(w**2)


def test_bug4():
    e = x/(w**4 + x**2*w**4 + 2*x*w**4)*w**4
    assert e.series(w, n=2).simplify() in [x/(1 + 2*x + x**2),
                                           1/(1 + x/2 + 1/x/2)/2, 1/x/(1 + 2/x + x**(-2))]


def test_bug5():
    e = (-log(w) + log(1 + w*log(x)))**(-2)*w**(-2)*((-log(w) +
                                                      log(1 + x*w))*(-log(w) + log(1 + w*log(x)))*w - x*(-log(w) +
                                                                                                         log(1 + w*log(x)))*w)
    assert e.series(w, n=0, logx=l) == (x/l + 1)/w + O(1, w)
    assert e.series(w, n=1, logx=l) == x*log(x)/l**2 + log(x)/l - \
        x/l + (1 + x/l)/w + O(w)


def test_sympyissue_4115():
    assert (sin(x)/(1 - cos(x))).series(x, n=1) == 2/x + O(x)
    assert (sin(x)**2/(1 - cos(x))).series(x, n=2) == 2 + O(x**2)


def test_pole():
    pytest.raises(PoleError, lambda: sin(1/x).series(x, 0, 5))
    pytest.raises(PoleError, lambda: sin(1 + 1/x).series(x, 0, 5))
    pytest.raises(PoleError, lambda: (x*sin(1/x)).series(x, 0, 5))
    pytest.raises(PoleError, lambda: besselk(0, x).series(x, 0, 2))


def test_expsinbug():
    assert exp(sin(x)).series(x, 0, 0) == O(1, x)
    assert exp(sin(x)).series(x, 0, 1) == 1 + O(x)
    assert exp(sin(x)).series(x, 0, 2) == 1 + x + O(x**2)
    assert exp(sin(x)).series(x, 0, 3) == 1 + x + x**2/2 + O(x**3)
    assert exp(sin(x)).series(x, 0, 4) == 1 + x + x**2/2 + O(x**4)
    assert exp(sin(x)).series(x, 0, 5) == 1 + x + x**2/2 - x**4/8 + O(x**5)


def test_floor():
    x = Symbol('x')
    assert floor(x).series(x) == 0
    assert floor(-x).series(x) == -1
    assert floor(sin(x)).series(x) == 0
    assert floor(sin(-x)).series(x) == -1
    assert floor(x**3).series(x) == 0
    assert floor(-x**3).series(x) == -1
    assert floor(cos(x)).series(x) == 0
    assert floor(cos(-x)).series(x) == 0
    assert floor(5 + sin(x)).series(x) == 5
    assert floor(5 + sin(-x)).series(x) == 4

    assert floor(x).series(x, 2) == 2
    assert floor(-x).series(x, 2) == -3

    x = Symbol('x', negative=True)
    assert floor(x + 1.5).series(x) == 1


def test_ceiling():
    assert ceiling(x).series(x) == 1
    assert ceiling(-x).series(x) == 0
    assert ceiling(sin(x)).series(x) == 1
    assert ceiling(sin(-x)).series(x) == 0
    assert ceiling(1 - cos(x)).series(x) == 1
    assert ceiling(1 - cos(-x)).series(x) == 1
    assert ceiling(x).series(x, 2) == 3
    assert ceiling(-x).series(x, 2) == -2


def test_abs():
    assert abs(x).series(x, n=4) == x
    assert abs(-x).series(x, n=4) == x
    assert abs(x + 1).series(x, n=4) == x + 1
    assert abs(sin(x)).series(x, n=4) == x - x**3/6 + O(x**4)
    assert abs(sin(-x)).series(x, n=4) == x - x**3/6 + O(x**4)
    assert abs(x - a).series(x, 1) == (x - a)/sign(1 - a)

    # issue sympy/sympy#5183
    assert abs(x + x**2).series(n=1) == O(x)
    assert abs(x + x**2).series(n=2) == x + O(x**2)


def test_dir():
    assert abs(x).series(x, 0, dir=-1) == x
    assert abs(x).series(x, 0, dir=+1) == -x
    assert floor(x + 2).series(x, 0, dir=-1) == 2
    assert floor(x + 2).series(x, 0, dir=+1) == 1
    assert floor(x + 2.2).series(x, 0, dir=+1) == 2
    assert ceiling(x + 2.2).series(x, 0, dir=+1) == 3
    assert sin(x + y).series(x, 0, dir=+1) == sin(x + y).series(x, 0, dir=-1)


def test_sympyissue_3504():
    e = asin(a*x)/x
    assert e.series(x, 4, n=2).removeO() == \
        (x - 4)*(a/(4*sqrt(-16*a**2 + 1)) - asin(4*a)/16) + asin(4*a)/4


def test_sympyissue_4441():
    f = 1/(1 + a*x)
    assert f.series(x, 0, 5) == 1 - a*x + a**2*x**2 - a**3*x**3 + \
        a**4*x**4 + O(x**5)
    f = 1/(1 + (a + b)*x)
    assert f.series(x, 0, 3) == 1 + x*(-a - b) + x**2*(a**2 + 2*a*b + b**2) + O(x**3)


def test_sympyissue_4329():
    assert tan(x).series(x, pi/2, n=3).removeO() == \
        -pi/6 + x/3 - 1/(x - pi/2)
    assert cot(x).series(x, pi, n=3).removeO() == \
        -x/3 + pi/3 + 1/(x - pi)
    assert limit(tan(x)**tan(2*x), x, pi/4) == exp(-1)


def test_sympyissue_5925():
    sx = sqrt(x + z).series(z, 0, 1)
    sxy = sqrt(x + y + z).series(z, 0, 1)
    s1, s2 = sx.subs({x: x + y}), sxy
    assert (s1 - s2).expand().removeO().simplify() == 0

    sx = sqrt(x + z).series(z, 0, 1)
    sxy = sqrt(x + y + z).series(z, 0, 1)
    assert sxy.subs({x: 1, y: 2}) == sx.subs({x: 3})


def test_sympyissue_6235():
    q = Symbol('q', positive=True)
    assert (((x - 1)**q + 1)/(x**q - 1)).nseries(x, n=2).removeO() == \
        (-1 - x**q + (-1)**(q + 1) + (-1)**(q + 1)*x**q +
         (-1)**q*q*x**(q + 1) + (-1)**q*q*x)


def test_sympyissue_6236():
    q = Symbol('q', positive=True)
    assert (((x - 1)**q)/(x**q - 1)).nseries(x, n=2).removeO() == \
        (-1)**(q + 1) + (-1)**(q + 1)*x**q + (-1)**q*q*x**(q + 1) + (-1)**q*q*x


def test_issue_210():
    assert cos(x**6).series(x, n=12) == 1 + O(x**12)
    assert cos(x**6).series(x, n=23) == 1 - x**12/2 + O(x**23)
    assert cos(x**6).series(x, n=24) == 1 - x**12/2 + O(x**24)
    assert cos(x**6).series(x, n=36) == 1 - x**12/2 + x**24/24 + O(x**36)

    # issue sympy/sympy#10503
    f = exp(x**3)*cos(x**6)
    assert f.series(x, n=14) == (1 + x**3 + x**6/2 +
                                 x**9/6 - 11*x**12/24 + O(x**14))
    assert f.series(x, n=15) == (1 + x**3 + x**6/2 +
                                 x**9/6 - 11*x**12/24 + O(x**15))
    assert f.series(x, n=16) == (1 + x**3 + x**6/2 + x**9/6 - 11*x**12/24 -
                                 59*x**15/120 + O(x**16))


def test_sympyissue_21075():
    e = (sqrt(x) + cbrt(x))**2
    assert e.series(x) == e.expand()


def test_sympyissue_21227():
    f = log(x)

    assert f.series(x, logx=y) == y
    assert f.series(x, logx=-x) == -x

    f = log(-log(x))

    assert f.series(x, logx=y) == log(-y)
    assert f.series(x, logx=-x) == log(x)

    f = log(log(x))

    assert f.series(x, logx=y) == log(y)
    assert f.series(x, logx=-x) == log(-x)
    assert f.series(x, logx=x) == log(x)



"""Configuration utilities."""

import ast
import contextlib
import os


__all__ = 'setup',


_default_config = {
    'USE_COLLINS_RESULTANT':      False,
    'USE_HEU_GCD':                True,
    'HEU_GCD_MAX':                6,

    'FALLBACK_GCD_ZZ_METHOD':     'prs',
    'GCD_AA_METHOD':              'prs',

    'USE_IRREDUCIBLE_IN_FACTOR':  False,
    'USE_CYCLOTOMIC_FACTOR':      True,

    'EEZ_RESTART_IF_NEEDED':      True,
    'EEZ_NUMBER_OF_CONFIGS':      3,
    'EEZ_NUMBER_OF_TRIES':        5,
    'EEZ_MODULUS_STEP':           2,

    'GF_IRRED_METHOD':            'rabin',
    'GF_FACTOR_METHOD':           'zassenhaus',

    'AA_FACTOR_METHOD':           'modular',

    'GROEBNER':                   'buchberger',
    'MINPOLY_METHOD':             'compose',

    'KARATSUBA_CUTOFF':           100,
}

_current_config = {}


@contextlib.contextmanager
def using(**kwargs):
    for k, v in kwargs.items():
        setup(k, v)

    yield

    for k in kwargs:
        setup(k)


def setup(key, value=None):
    """Assign a value to (or reset) a configuration item."""
    key = key.upper()

    if value is not None:
        _current_config[key] = value
    else:
        _current_config[key] = _default_config[key]


def query(key):
    """Ask for a value of the given configuration item."""
    return _current_config.get(key.upper(), None)


def configure():
    """Initialized configuration of polys module."""
    for key, default in _default_config.items():
        _current_config[key] = default

        value = os.getenv('DIOFANT_' + key)
        if value is not None:
            try:
                value = ast.literal_eval(value)
            except (SyntaxError, ValueError):
                pass
            else:
                if type(value) is type(default):
                    _current_config[key] = value


configure()
